(** Collection of composable representations *)

open Repr.Composable

let eprintf s = Peahell.Report.fprintf Fmt.stderr s
let index_of_constr l k =
  match List.concat (List.mapi (fun i (k',_) -> if k' = k then [i] else []) l) with
  | [i] -> i
  | _ ->
    eprintf "Could not find constructor %s in type %a@." k Types.pp_evaled (Types.Sum l);
    assert false

module BoxedIngredients = struct
  let recompose ty_env ty_sym (h : Hole.Head.t) =
    match Name.Map.find ty_sym ty_env, h with
    | _, Hole -> Mem.Hole.(MHole WordOps.Id)
    | Types.Ref _, Ref ->
      let hole = Mem.Hole.(MHole WordOps.Id) in
      Mem.Hole.MHPtr { len = 64; align = 0; hole }
    | Types.Product _, Tuple { arity; pos } ->
      let hole = Mem.Hole.(MHole WordOps.Id) in
      let block = Mem.Hole.MHBlock { pos; hole } in
      Mem.Hole.MHPtr { len = 64; align = 0; hole = block }
    | Types.Sum _, Constr { constr; arity; pos } ->
      let hole = Mem.Hole.(MHole WordOps.Id) in
      let block = Mem.Hole.MHBlock { pos = pos + 1; hole } in
      Mem.Hole.MHPtr { len = 64; align = 0; hole = block }
    | ty, _ ->
      Fmt.invalid_arg "Unsupported combination:@.%a,@.%a"
        Types.pp_evaled ty
        Hole.Head.pp h

  let tagsets_of_word wspec = [ Tag.Klass.starter_ints wspec ]
  let discrs_of_word wspec =
    let tagset = Tag.Klass.starter_ints wspec in
    let split = Tag.KlassMap.(add tagset None empty_intm) in
    [ { position = Mem.Hole.(MHole WordOps.Id); min_width = Some 64; split } ]

  let discrs_of_sum ty_env sumspec =
    let position = Mem.Hole.MHPtr { len = 64; align = 0;
      hole = Mem.Hole.MHBlock { pos = 0; hole = Mem.Hole.(MHole WordOps.Id) }
    }
    and split = List.fold_left
      (fun acc (constr, _) -> Tag.KlassMap.add
        (Tag.Klass.singleton_constrs constr)
        (Some (Mem.Word (Z.of_int @@ index_of_constr sumspec constr)))
        acc)
      Tag.KlassMap.empty_constrm sumspec in
    [ { position; min_width = Some 64; split } ]

end
module Boxed = Make(BoxedIngredients)

module NicheIngredients = struct
  let recompose ty_env ty_sym (h : Hole.Head.t) =
    match Name.Map.find ty_sym ty_env, h with
    | Types.Sum [ k0, []; k1, [_] ],
      Constr { constr = k; arity = 1; pos = 0 }
      when k = k1 ->
      let hole = Mem.Hole.(MHole WordOps.Id) in
      Mem.Hole.MHPtr { len = 64; align = 1; hole }
    | _ -> BoxedIngredients.recompose ty_env ty_sym h
  
  let tagsets_of_word = BoxedIngredients.tagsets_of_word
  let discrs_of_word = BoxedIngredients.discrs_of_word

  let discrs_of_sum ty_env sumspec = match sumspec with
    | [ k0, []; k1, [_] ] ->
      let ts0 = Tag.Klass.singleton_constrs k0
      and ts1 = Tag.Klass.singleton_constrs k1 in
      let split = Tag.KlassMap.(add ts1 (Some Mem.True) @@
        add ts0 (Some Mem.False) empty_constrm) in
      [ { position = Mem.Hole.(MHole WordOps.Id); min_width = None; split } ]
    | _ -> BoxedIngredients.discrs_of_sum ty_env sumspec

end
module Niche = Make(NicheIngredients)

module CamlIngredients = struct
  let rec recompose (ty_env : Types.evaled_env) ty_sym (h : Hole.Head.t) =
    match Name.Map.find ty_sym ty_env, h with
    | _, Hole -> Mem.Hole.(MHole WordOps.Id)
    | Types.Var ty_sym, _ -> recompose ty_env ty_sym h
    | Types.Product _, Tuple { arity; pos } ->
      let hole = Mem.Hole.(MHole WordOps.Id) in
      Mem.Hole.MHPtr { len = 64; align = 1;
                       hole = Mem.Hole.MHBlock { pos = pos + 1; hole }
                     }
    | Types.Sum _, Constr { constr; arity; pos } ->
      let hole = Mem.Hole.(MHole WordOps.Id) in
      Mem.Hole.MHPtr { len = 64; align = 1;
                       hole = Mem.Hole.MHBlock { pos = pos + 1; hole }
                     }
    | ty, _ ->
      Fmt.invalid_arg "Unsupported combination:@.%a,@.%a"
        Types.pp_evaled ty
        Hole.Head.pp h

  let partition_constrs l =
    List.fold_left
      (fun (unit, nonunit) (k,l) -> match l with
        | [] -> (Tag.Klass.add k unit, nonunit)
        | _ -> (unit, Tag.Klass.add k nonunit))
      (Tag.Klass.empty_constrs, Tag.Klass.empty_constrs) l

  let shift_imm z = Z.logor (Z.shift_left z 1) Z.one

  let tagsets_of_word wspec = [ Tag.Klass.starter_ints wspec ]
  let discrs_of_word wspec =
    let tagset = Tag.Klass.starter_ints wspec in
    let split = Tag.KlassMap.(add tagset None empty_intm)
    and untangle_tag =
      Mem.Hole.WordOps.(Arith (Arith (Id, Sub, Z.one), Div, Z.(one + one))) in
    [ { position = Mem.Hole.MHole untangle_tag; min_width = Some 64; split } ]

  let discrs_of_sum ty_env sumspec =
    let unitl, nonunitl = List.partition (fun (_, l) -> l = []) sumspec in
    let unit, nonunit =
      Tag.Klass.starter_constrs unitl, Tag.Klass.starter_constrs nonunitl in
    let top_discr = {
      position = Mem.Hole.(MHole
        WordOps.(Cmp (Extract { low = 0; high = 0; ops = Id }, Neq, Z.zero)));
      min_width = Some 64;
      split = Tag.KlassMap.(add unit (Some Mem.True) @@
        add nonunit (Some Mem.False) empty_constrm)
    }
    and unit_discr = {
      position = Mem.Hole.(MHole WordOps.Id);
      min_width = Some 64;
      split = List.fold_left
        (fun acc (constr, _) ->
          let z = shift_imm @@ Z.of_int @@ index_of_constr sumspec constr in
          Tag.KlassMap.add (Tag.Klass.singleton_constrs constr) (Some (Mem.Word z)) acc)
        Tag.KlassMap.empty_constrm unitl
    }
    and nonunit_discr = {
      position = Mem.Hole.(MHPtr { len = 64; align = 1;
        hole = MHBlock { pos = 0; hole = (MHole WordOps.Id) } });
      min_width = Some 64;
      split = List.fold_left
        (fun acc (constr, _) ->
          let z = Z.of_int @@ index_of_constr sumspec constr in
          Tag.KlassMap.add (Tag.Klass.singleton_constrs constr) (Some (Mem.Word z)) acc)
        Tag.KlassMap.empty_constrm nonunitl
    } in
    [ top_discr; unit_discr; nonunit_discr ]
  
end
module Caml = Make(CamlIngredients)

(*module ScalaIngredients = struct
  let klass_word ty k = "Class" ^ k (* TODO *)

  let tup_subword_width ty pos = if pos = 0 then 32 else 64 (* TODO *)
  let constr_subword_width ty k pos = if pos = 0 then 32 else 64 (* TODO *)
  let tup_field_offset ty pos = pos + 1, 0 (* TODO *)
  let constr_field_offset ty k pos = pos + 1, 0 (* TODO *)

  let block_nest pos offset align mh =
    let inner_block = Mem.Hole.MHBlock { align; pos = offset; hole = mh } in
    let inner_block = if (pos = 0) (* subword next to klass is actually a subsubword *)
      then Mem.Hole.MHBlock { align = 32; pos = 1; hole = inner_block }
      else inner_block
    in
    let outer_block = Mem.Hole.MHBlock { align = 64; pos = pos + 1; hole = inner_block } in
    Mem.Hole.MHPtr { len = 32; align = 0; hole = outer_block }

  let recompose ty (h : Hole.Head.t) = match Types.unroll ty, h with
    | _, Hole -> Mem.Hole.(MHole WordOps.Id)
    | Types.Product l, Tuple { pos; _ } ->
      let mh = Mem.Hole.(MHole WordOps.Id) in
      let k, j = tup_field_offset ty pos in
      let a = tup_subword_width ty k in
      block_nest k j a mh
    | Types.Sum l, Constr { constr; pos; _ } ->
      let mh = Mem.Hole.(MHole WordOps.Id) in
      let k, j = constr_field_offset ty constr pos in
      let a = constr_subword_width ty constr pos in
      block_nest k j a mh
    | _ -> assert false

  let g ty tag = match Types.unroll ty, tag with
    | Types.Word _, Tag.Int c -> Mem.Pattern.Imm (Word c) (* TODO width *)
    | Types.Sum l, Tag.Constr k -> Mem.Pattern.Imm (Symbol (klass_word ty k)) (* TODO width = 32 *)
    | _ -> assert false

  let h ty = match Types.unroll ty with
    | Types.Word _ -> Mem.Hole.(MHole WordOps.Id)
    | Types.Sum _ ->
      Mem.Hole.MHPtr {
        len = 32; align = 0;
        hole = Mem.Hole.MHBlock {
            align = 64; pos = 1;
            hole = Mem.Hole.MHBlock { align = 32; pos = 0; hole = (MHole WordOps.Id) }
          }
      }
    | _ -> assert false
end
module Scala = MakeSimple(ScalaIngredients)*)
