open Types
open Repr.Composable

let eprintf s = Peahell.Report.fprintf Fmt.stderr s
let index_of_constr l k =
  match List.concat (List.mapi (fun i (k',_) -> if k' = k then [i] else []) l) with
  | [i] -> i
  | _ ->
    eprintf "Could not find constructor %s in type %a@." k Types.pp_evaled (Types.Sum l);
    assert false

let ptr_width = 64
let two = Z.(one + one)

module Ingredients = struct
  let sum_ints_of_elts f l =
    List.fold_left (fun acc x -> acc + (f x)) 0 l
  let sum_zs_of_elts f l =
    List.fold_left (fun acc x -> Z.(acc + (f x))) Z.zero l

  let rec ty_width (env : Types.evaled_env) sym =
    match Types.find_and_unroll sym env with
    | Word { width; _ } -> width
    | Ref _ -> ptr_width
    | Product l -> sum_ints_of_elts (ty_width env) l
    | Sum l ->
      let subsum = sum_ints_of_elts (ty_width env) in
      let cardsum = List.fold_left
        (fun acc (_, syms) -> Z.(acc + two ** (subsum syms)))
        Z.zero l in
      Z.log2up cardsum
    | _ -> assert false

  let rec get_constr_offsets ty_env cur_aoff l = match l with
    | [] -> []
    | (constr, tys) :: l ->
      let nb_vals = Z.(two ** (sum_ints_of_elts (ty_width ty_env) tys)) in
      let next_aoff = Z.(cur_aoff + nb_vals) in
      let pat = cur_aoff, Z.(next_aoff - one) in
      (constr, pat) :: get_constr_offsets ty_env next_aoff l

  let rec recompose (ty_env : Types.evaled_env) ty_sym h =
    match Types.find_and_unroll ty_sym ty_env, h with
    | _, Hole.Head.Hole -> Mem.Hole.(MHole WordOps.Id)
    | Types.Ref _, Hole.Head.Ref ->
      let hole = Mem.Hole.(MHole WordOps.Id)in
      Mem.Hole.MHPtr { len = ptr_width; align = 0; hole }
    | Types.Product l, Hole.Head.Tuple { arity; pos } ->
      let high = sum_ints_of_elts (ty_width ty_env) (List.filteri (fun i _ -> i > pos) l)
      and low = sum_ints_of_elts (ty_width ty_env) (List.filteri (fun i _ -> i < pos) l) in
      Mem.Hole.(MHole (Extract { high; low; ops = Id }))
    | Types.Sum l, Hole.Head.Constr { constr; arity; pos } ->
      let syms = List.assoc constr l in
      let arith_offset, _ = List.assoc constr (get_constr_offsets ty_env Z.zero l) in
      let low =
        sum_ints_of_elts (ty_width ty_env) (List.filteri (fun i _ -> i < pos) syms) in
      let high = low + (ty_width ty_env (List.nth syms pos)) - 1 in
      Mem.Hole.(MHole (Extract { high; low; ops = Arith (Id, Sub, arith_offset) }))
    | ty, _ ->
      Fmt.invalid_arg "Unsupported combination:@.%a,@.%a"
        Types.pp_evaled ty
        Hole.Head.pp h

  let tagsets_of_word wspec = [ Tag.Klass.starter_ints wspec ]
  let discrs_of_word wspec =
    let tagset = Tag.Klass.starter_ints wspec in
    let split = Tag.KlassMap.(add tagset None empty_intm) in
    [ { position = Mem.Hole.(MHole WordOps.Id); min_width = None; split } ]

  let discrs_of_sum ty_env sumspec =
    let discr_of_constr constr offset next_constrs =
      let tagset = Tag.Klass.singleton_constrs constr in
      let split = Tag.KlassMap.(add tagset (Some Mem.True) @@
        List.fold_left
          (fun acc (k, _) -> add (Tag.Klass.singleton_constrs k) (Some Mem.False) acc)
          empty_constrm next_constrs)
      and position = Mem.Hole.(MHole WordOps.(Cmp (Id, Leq, offset))) in
      { position; min_width = None; split } in
    let rec aux constrs offsets = match constrs, offsets with
      | [], [] -> []
      | (constr, _) :: constrs, (_, offset) :: offsets ->
        discr_of_constr constr offset constrs :: aux constrs offsets
      | _ -> assert false in
    let _, offsets = List.split (get_constr_offsets ty_env Z.zero sumspec) in
    aux sumspec offsets

end

include Make(Ingredients)
