open Z3

let out_sym = "y"
let arg_sym h = Fmt.str "x%a" Hole.pp h

type sorted_expr = { sort : Sort.sort; expr : Expr.expr }
type params = (Hole.t * sorted_expr) list
type decl =
  | Cst of Expr.expr
  | Func of FuncDecl.func_decl
type problem = { declaration : decl; assertion : Expr.expr }
type interp = { cases : (Expr.expr list * Expr.expr) list; default : Expr.expr }
type solution = params * interp

let pp_expr fmt expr =
  Fmt.pf fmt "%s" (Expr.to_string expr)
let pp_decl fmt = function
  | Cst e ->
    Fmt.pf fmt "(define-fun %s () %s)"
      (Expr.to_string e) (Sort.to_string @@ Expr.get_sort e)
  | Func fd ->
    Fmt.pf fmt "%s" (FuncDecl.to_string fd)
let pp_problem fmt { declaration; assertion } =
  Fmt.pf fmt "%a@.(assert %a)@." pp_decl declaration pp_expr assertion
let pp_interp fmt = function
  | { cases = []; default } -> pp_expr fmt default
  | { cases; default } ->
    Fmt.pf fmt "%a; _ -> %a"
      Fmt.(list ~sep:semi
        (pair ~sep:(any " -> ") (list ~sep:comma pp_expr) pp_expr))
        cases
      pp_expr default

let sort_of = Expr.get_sort

let eval_expr (eval_bool : bool -> 'a) (eval_bv : Z.t -> 'a) : Expr.expr -> 'a =
  fun expr -> match Sort.get_sort_kind (sort_of expr) with
  | BV_SORT ->
    (try eval_bv @@ Z.of_string @@ BitVector.numeral_to_string expr
    with Error _ ->
      Peahell.Report.fprintf Fmt.stderr "Could not eval expr %a" pp_expr expr;
      assert false)
  | BOOL_SORT -> (match Boolean.get_bool_value expr with
    | L_FALSE -> eval_bool false
    | L_TRUE -> eval_bool true
    | L_UNDEF -> assert false)
  | _ -> assert false

let rec eval_solution (eval_param : Hole.t -> solution) (params, { cases; default }) =
  let simple_eval_expr =
    eval_expr (function true -> Z.one | false -> Z.zero) (fun z -> z)
  in match params with
  | [] -> default
  | params ->
    let args = List.map (fun (h, _) -> eval_solution eval_param @@ eval_param h) params in
    match List.find_opt
      (fun (param_vals, _) ->
        List.for_all2 (fun arg pv -> simple_eval_expr arg = simple_eval_expr pv)
          args param_vals)
      cases with
    | Some (_, res) -> res
    | None -> default

let ctx = mk_context []
let solver = Solver.mk_solver ctx None

let solve_problem ~concrete { declaration; assertion } : interp =
  match Solver.(reset solver; add solver [assertion]; check solver []) with
  | SATISFIABLE -> (match Solver.get_model solver with
    | Some model ->
      (match declaration with
      | Cst expr -> (match Model.eval model expr concrete with
        | Some e -> { cases = []; default = e }
        | None -> failwith "no value")
      | Func decl -> match Model.get_func_interp model decl with
        | Some fi -> Model.FuncInterp.{
            cases = List.map
              (fun ent -> FuncEntry.(get_args ent, get_value ent))
              (get_entries fi);
            default = get_else fi }
        | None -> failwith "no repr func interpretation")
    | None -> failwith "no model")
  | UNSATISFIABLE -> failwith "UNSAT"
  | UNKNOWN -> failwith "Dunno"

let bv_sort width = BitVector.mk_sort ctx width
let width_of_bv_sort sort =
  match Sort.get_sort_kind sort with
  | BV_SORT -> BitVector.get_size sort
  | _ -> failwith "not a bitvector sort"

let bv_ule = BitVector.mk_ule ctx
let bv_ult = BitVector.mk_ult ctx
let bv_uge = BitVector.mk_uge ctx
let bv_ugt = BitVector.mk_ugt ctx

let bv_add = BitVector.mk_add ctx
let bv_sub = BitVector.mk_sub ctx
let bv_mul = BitVector.mk_mul ctx
let bv_div = BitVector.mk_udiv ctx
let bv_mod = BitVector.mk_smod ctx

let bv_extract = BitVector.mk_extract ctx

let bool_sort = Boolean.mk_sort ctx

let bool_true = Boolean.mk_true ctx
let bool_false = Boolean.mk_false ctx

let bool_not = Boolean.mk_not ctx
let bool_and e1 e2 = Boolean.mk_and ctx [ e1; e2 ]

let string_of_expr = Z3.Expr.to_string

let eq_expr = Boolean.mk_eq ctx
let neq_expr e e' = Boolean.mk_not ctx @@ eq_expr e e'

let expr_of_z sort z = Expr.mk_numeral_string ctx (Z.to_string z) sort
let expr_of_zero sort = Expr.mk_numeral_string ctx "0" sort

let const_expr sort sym = Expr.mk_const_s ctx sym sort

let forall_expr args body =
  Quantifier.(expr_of_quantifier @@ mk_forall_const ctx args body None [] [] None None)

let encode sort params (apply : params -> sorted_expr -> sorted_expr) : problem =
  match params with
  | [] ->
    let expr = const_expr sort out_sym in
    let app = apply [] { sort; expr } in
    { declaration = Cst expr; assertion = app.expr }
  | _ ->
    let param_sorts, param_exprs = List.split @@
      List.map (fun (_, { sort; expr }) -> sort, expr) params in
    let decl = FuncDecl.mk_func_decl_s ctx out_sym param_sorts sort in
    let expr = FuncDecl.apply decl param_exprs in
    let app = apply params { sort; expr } in
    { declaration = Func decl; assertion = forall_expr param_exprs app.expr }
