let two = Z.(one + one)

module IntervalTree = struct
  include Diet.Make(Z)

  let interval x y = add (Interval.make x y) empty
  let singleton x = interval x x

  let remove_point x set = remove (Interval.make x x) set
  let add_point x set = add (Interval.make x x) set

  let signed width =
    let e = width - 1 in
    interval Z.(~- two ** e) Z.(two ** e - one)

  let unsigned width =
    let e = width in
    interval Z.zero Z.(two ** e - one)

  let pick_one it = Interval.x (choose it)

  let arith_shift z it =
    fold (fun i acc -> add Interval.(make Z.(x i + z) Z.(y i + z)) acc) it empty

  let wordsize = 64
  let address = remove_point Z.zero (unsigned 64)

  let pp_interval fmt i =
    let x = Interval.x i and y = Interval.y i in
    if Z.equal x y then
      Fmt.pf fmt "%a" Z.pp_print x
    else
      Fmt.pf fmt "%a--%a" Z.pp_print x Z.pp_print y
  let pp fmt values =
    Fmt.(iter ~sep:(any " &@ ")) iter pp_interval fmt values
end

module BitWord = struct
  type t = (bool option) list

  let zeroes l = List.init l (fun _ -> Some false)
  let ones l = List.init l (fun _ -> Some true)
  let alls l = List.init l (fun _ -> None)

  let fold_sum_bits f =
    List.fold_left
      (fun (acc, twop) b -> (if (f b) then Z.(acc + twop) else acc), Z.(twop * two))
      (Z.zero, Z.one)

  let staged_mem e l =
    let zero_mask, _ = fold_sum_bits ((=)(Some false)) l
    and one_mask, _ = fold_sum_bits ((=)(Some true)) l in
    let zero_cond = Target.(RNot (RAnd (e, zero_mask)))
    and one_cond = Target.(RCmp (RAnd (e, one_mask), Geq, one_mask)) in
    Target.RLogAnd (zero_cond, one_cond)

  let concat_le = List.concat

  let cardinal =
    List.fold_left (fun acc -> function None -> Z.(acc * two) | _ -> acc) Z.one

  let minimum l =
    let res, _ = fold_sum_bits ((=)(Some true)) l in res

  let rec mem z = function
    | [] -> Z.(z = zero)
    | None :: l -> mem Z.(z asr 1) l
    | (Some b) :: l -> ((Z.testbit z 0) = b) && (mem Z.(z asr 1) l)

  let rec is_subset_of l1 l2 = match l1, l2 with
    | [], _ -> true
    | _, [] -> false
    | b1 :: l1, b2 :: l2 -> (match b1, b2 with
      | _, None -> is_subset_of l1 l2
      | Some b1, Some b2 -> (b1 = b2) && (is_subset_of l1 l2)
      | _ -> false)

  let rec union l1 l2 = match l1, l2 with
    | [], l | l, [] -> union [Some false] l
    | b1 :: l1, b2 :: l2 ->
      (if (b1 = b2) then b1 else None) :: (union l1 l2)

  let rec inter l1 l2 = match l1, l2 with
    | [], l | l, [] -> []
    | None :: l1, b :: l2 | b :: l1, None :: l2 -> b :: (inter l1 l2)
    | (Some b1) :: l1, (Some b2) :: l2 ->
      if (b1 = b2) then (Some b1) :: (inter l1 l2) else assert false

  let pp fmt =
    let pp_bit fmt bit = Fmt.pf fmt "%c" (match bit with
      | None -> '_'
      | Some false -> '0'
      | Some true -> '1')
    in
    Fmt.(list pp_bit) fmt
end

type valt = Z.t
type width = int
type t = Empty of width | ITree of width * IntervalTree.t | BWord of BitWord.t

(** Technically always possible but way too expensive after X intervals *)
let bword_to_itree bw =
  let max_splits = 10 in
  let rec aux splits aoff bw =
    if splits > max_splits then None
    else match bw with
    | [] -> Some (IntervalTree.singleton aoff)
    | (Some false) :: bw -> aux splits aoff bw
    | (Some true) :: bw ->
      let wid = List.length bw in
      aux splits Z.(aoff + two ** wid) bw
    | None :: bw ->
      let wid = List.length bw in
      let it0 = aux (splits + 1) aoff bw
      and it1 = aux (splits + 1) Z.(aoff + two ** wid) bw in
      (match it0, it1 with
       | Some it0, Some it1 -> Some (IntervalTree.union it0 it1)
       | _ -> None)
  in
  aux 0 Z.zero bw

(** Not always possible *)
let itree_to_bword l it = (* TODO *) assert false

let staged_mem s e =
  let res = match s with
  | Empty _ -> Target.RWord Z.zero
  | ITree (l, it) ->
    let aux i =
      let x, y = IntervalTree.Interval.(x i, y i) in
      Target.RLogAnd ((RCmp (e, Geq, x)), (RCmp (e, Leq, y)))
    in
    IntervalTree.fold
      (fun i acc -> Target.RLogAnd(acc, aux i)) it (Target.RWord Z.one)
  | BWord bw -> BitWord.staged_mem e bw
  in
  Target.simplify_rexpr res

let empty w = Empty w

let singleton l z = ITree (l, IntervalTree.singleton z)
let range l x y = ITree (l, IntervalTree.interval x y)
let full_range l = (* could also be a bitword *)
  ITree (l, IntervalTree.interval Z.zero Z.(two ** l - one))
let non_zero l =
  ITree (l, IntervalTree.interval Z.one Z.(two ** l - one))

let subrange l start stop =
  BWord BitWord.(concat_le [zeroes start; alls (stop - start); ones (l - stop)])
let aligned_address l a = BWord BitWord.(concat_le [zeroes a; alls (l - a)])

let is_empty = function
  | Empty _ -> true
  | ITree (l, it) -> (l = 0) || (IntervalTree.is_empty it)
  | BWord bw -> (bw = [])
let cardinal = function
  | Empty _ -> Some Z.zero
  | ITree (_, it) -> Some (IntervalTree.cardinal it)
  | BWord bw -> Some (BitWord.cardinal bw)

let pick_one = function
  | Empty _ -> assert false
  | ITree (_, it) -> IntervalTree.pick_one it
  | BWord bw -> BitWord.minimum bw

let mem z s = match s with
  | Empty _ -> false
  | ITree (_, it) -> IntervalTree.mem z it
  | BWord bw -> BitWord.mem z bw

let is_subset_of s1 s2 = match s1, s2 with
  | Empty _, _ -> true
  | _, Empty _ -> false
  | ITree (_, it1), ITree (_, it2) ->
    (IntervalTree.equal it1 it2) || ((IntervalTree.compare it1 it2) < 0)
  | BWord bw1, BWord bw2 -> BitWord.is_subset_of bw1 bw2
  | _ -> assert false

let union s1 s2 = match s1, s2 with
  | Empty _, s | s, Empty _ -> s
  | ITree (l1, it1), ITree (l2, it2) ->
    ITree ((if l1 >= l2 then l1 else l2), IntervalTree.union it1 it2)
  | BWord bw1, BWord bw2 -> BWord (BitWord.union bw1 bw2)
  | BWord bw, ITree (l, it) | ITree (l, it), BWord bw ->
    (match itree_to_bword l it with
     | None -> (match bword_to_itree bw with
                | None -> assert false
                | Some it' ->
                  let l = (if l > (List.length bw) then l else List.length bw) in
                  ITree (l, IntervalTree.union it it'))
     | Some bw' -> BWord (BitWord.union bw bw'))

let inter s1 s2 = match s1, s2 with
  | Empty w, _ | _, Empty w -> Empty w (* TODO Compute right width *)
  | ITree (l1, it1), ITree (l2, it2) ->
    ITree ((if l1 <= l2 then l1 else l2), IntervalTree.inter it1 it2)
  | BWord bw1, BWord bw2 -> BWord (BitWord.inter bw1 bw2)
  | BWord bw, ITree (l, it) | ITree (l, it), BWord bw ->
    (match itree_to_bword l it with
     | None -> (match bword_to_itree bw with
                | None -> assert false
                | Some it' ->
                  let l = (if l > (List.length bw) then l else List.length bw) in
                  ITree (l, IntervalTree.inter it it'))
     | Some bw' -> BWord (BitWord.inter bw bw'))

let rec complement = function
  | Empty w -> BWord (BitWord.ones w)
  | ITree (l, it) ->
    ITree (l, IntervalTree.(diff (interval Z.zero Z.(two ** l - one)) it))
  | BWord bw -> (match bword_to_itree bw with
      | None -> assert false
      | Some it -> complement (ITree (List.length bw, it)))

let rec diff s1 s2 = match s1, s2 with
  | Empty w, _ -> Empty w (* TODO Compute right width *)
  | s, Empty _ -> s (* TODO Compute right width *)
  | ITree (l1, it1), ITree (l2, it2) -> ITree (l1, IntervalTree.diff it1 it2)
  | BWord bw1, BWord bw2 ->
    (match bword_to_itree bw1, bword_to_itree bw2 with
     | Some it1, Some it2 ->
       diff (ITree (List.length bw1, it1)) (ITree (List.length bw2, it2))
     | _ -> assert false)
  | BWord bw, ITree (l, it) | ITree (l, it), BWord bw ->
    (match bword_to_itree bw with
     | Some it' -> diff (ITree (l, it)) (ITree (List.length bw, it'))
     | None -> assert false)

let rec arith_shift z = function
  | Empty w -> Empty w
  | ITree (l, it) -> ITree (l, IntervalTree.arith_shift z it)
  | BWord bw -> (match bword_to_itree bw with
                 | Some it -> arith_shift z (ITree (List.length bw, it))
                 | None -> assert false)

let rec slice start stop = function
  | Empty w -> Empty w
  | ITree (l, it) -> (match itree_to_bword l it with
    | Some bw -> slice start stop (BWord bw)
    | None -> assert false)
  | BWord bw -> BWord (List.filteri (fun i _ -> start <= i && i < stop) bw)

let pp fmt = function
  | Empty _ -> Fmt.pf fmt "%s" "{}"
  | ITree (l, it) -> Fmt.pf fmt "%a on %u bits" IntervalTree.pp it l
  | BWord bw -> Fmt.pf fmt "%a" BitWord.pp bw

(*type t = Empty | Singleton of widtht * valt | Wildcard of widtht
       | Neg of t | Concat of t * t | Union of t * t | Inter of t * t

let rec width_of = function
  | Empty -> assert false
  | Singleton (w, _) -> w
  | Wildcard w -> w
  | Neg s -> width_of s
  | Concat (s1, s2) -> (width_of s1) + (width_of s2)
  | Union (s1, s2) | Inter (s1, s2) -> assert (width_of s1 = width_of s2); width_of s1

let rec narrow start width = function
  | Empty -> Empty
  | Singleton (w, z) -> Singleton (width, Z.((z asr start) mod (of_int width)))
  | Wildcard _ -> Wildcard width
  | Neg s -> Neg (narrow start width s)
  | Concat (s1, s2) when width_of s1 <= start -> narrow (start - (width_of s1)) width s2
  | Concat (s1, s2) when width_of s1 >= start + width -> narrow start width s1
  | Concat (s1, s2) -> 
    let w1, w2 = width_of s1, width_of s2 in
    let s1 = narrow start (w1 - start) s1 in
    let s2 = narrow 0 (width - w1 + start) s2 in 
    Concat (s1, s2)
  | Union (s1, s2) -> Union (narrow start width s1, narrow start width s2)
  | Inter (s1, s2) -> Inter (narrow start width s1, narrow start width s2)

let rec eval_neg = function
  | Empty -> Wildcard (assert false)
  | Wildcard _ -> Empty
  | Singleton (w, z) -> Neg (Singleton (w, z))
  | Neg s -> s
  | Union (s1, s2) -> Inter (eval_neg s1, eval_neg s2)
  | Inter (s1, s2) -> Union (eval_neg s1, eval_neg s2)
  | Concat (s1, s2) ->
    let ns1, ns2 = eval_neg s1, eval_neg s2 in
    Union (Concat (ns1, s2), Union (Concat (s1, ns2), Concat (ns1, ns2)))

let non_zero len = Neg (Singleton (len, Z.zero))

let aligned_address len align = (* big-endian *)
  Concat (Wildcard (len - align), Singleton (align, Z.zero))

let empty = Empty

let singleton width z = Singleton (width, z)

let subrange width start stop =
  Concat (Singleton (width - stop, Z.zero), Concat (Wildcard (stop - start),
          Singleton (start, Z.(two ** start - one))))

let full_range width = Wildcard width

let rec is_empty = function (* TODO complete Inter case? *)
  | Empty -> true
  | Wildcard _ | Singleton _ -> false (* TODO deal with zero-width sets *)
  | Neg s -> is_empty (eval_neg s)
  | Concat (s1, s2) -> is_empty s1 || is_empty s2
  | Union (s1, s2) -> is_empty s1 && is_empty s2
  | Inter (s1, s2) -> is_empty s1 || is_empty s2 (* fuck it *)

let rec cardinal = function
  | Empty -> Some Z.zero
  | Singleton _ -> Some Z.one
  | Wildcard w -> Some Z.(two ** w)
  | Neg s ->
    (match cardinal s with
     | Some z -> let w = width_of s in Some Z.(two ** w - z)
     | None -> None)
  | Concat (s1, s2) ->
    (match cardinal s1, cardinal s2 with
     | Some z1, Some z2 -> Some Z.(z1 * z2)
     | _ -> None)
  | Union (s1, s2) -> None
  | Inter (s1, s2) -> None

let rec pick_one = function
  | Empty -> assert false
  | Singleton (_, z) -> z
  | Wildcard _ -> Z.zero
  | Neg s -> assert false (* TODO *)
  | Concat (s1, s2) ->
    let z1, z2 = pick_one s1, pick_one s2 in
    let w = width_of s2 in Z.(z1 lsl w + z2)
  | Union (s1, s2) -> pick_one s1
  | Inter (s1, s2) -> assert false (* TODO *)

let rec mem x s = match s with
  | Empty -> false
  | Singleton (_, z) -> x = z
  | Wildcard _ -> true (* TODO maybe check width? *)
  | Neg s -> not (mem x s)
  | Concat (s1, s2) ->
    let w = width_of s2 in
    let x1 = Z.(x asr w) in let x2 = Z.(x - x1) in
    (mem x1 s1) && (mem x2 s2)
  | Union (s1, s2) -> (mem x s1) || (mem x s2)
  | Inter (s1, s2) -> (mem x s1) && (mem x s2)

let staged_mem s e =
  let rec aux s e = match s with
    | Empty -> Target.RWord Z.zero
    | Wildcard _ -> Target.RWord Z.one
    | Singleton (w, z) -> Target.(RNot (RXor (e, z)))
    | Neg s -> Target.RNot (aux s e)
    | Concat (s1, s2) ->
      let w = width_of s2 in
      let e1 = Target.RRShift (e, w)
      and e2 = Target.(RAnd (e, Z.(two ** w - one))) in
      Target.RLogAnd (aux s1 e1, aux s2 e2)
    | Union (s1, s2) -> Target.RLogAnd (aux s1 e, aux s2 e)
    | Inter (s1, s2) -> Target.RLogOr (aux s1 e, aux s2 e)
  in
  Target.simplify_rexpr (aux s e)

let rec is_subset_of sub set = (sub = set) || match sub, set with
  | Empty, _ -> true
  | _, Empty -> false
  | Singleton (w, z), s when width_of s = w -> mem z s
  | s, Wildcard w when width_of s = w -> true
  | Union (s1, s2), _ -> is_subset_of s1 set && is_subset_of s2 set
  | sub, Concat (s1, s2) ->
    let sub1 = narrow 0 (width_of s1) sub
    and sub2 = narrow (width_of s1) (width_of s2) sub in
    is_subset_of sub1 set && is_subset_of sub2 set
  (* correctness is only important when set = aligned_address _ *)
  | Inter (s1, s2), _ -> is_subset_of s1 set || is_subset_of s2 set
  | _ -> false (* TODO *)

let union s1 s2 = match s1, s2 with
  | Empty, s | s, Empty -> s
  | Wildcard w, _ | _, Wildcard w -> Wildcard w
  | _ -> Union (s1, s2) (* TODO *)

let inter s1 s2 = match s1, s2 with
  | Empty, _ | _, Empty -> Empty
  | Wildcard _, s | s, Wildcard _ -> s
  | Singleton (w, z), s | s, Singleton (w, z) ->
    if mem z s then Singleton (w, z) else Empty
  | _ -> Inter (s1, s2) (* TODO *)

let diff s1 s2 = match s1, s2 with
  | Empty, _ -> Empty
  | s, Empty -> s
  | _ -> Inter (s1, Neg s2) (* TODO *)

let complement = eval_neg

let add_point width x s = union (singleton width x) s
let remove_point width x s = diff s (singleton width x)

let rec arith_shift z = function
  | Singleton (w, z') -> Singleton (w, Z.(z' + z))
  | Neg s -> Neg (arith_shift z s)
  | Concat (s1, s2) -> assert false (* TODO *)
  | Union (s1, s2) -> Union (arith_shift z s1, arith_shift z s2)
  | Inter (s1, s2) -> Inter (arith_shift z s1, arith_shift z s2)
  | s -> s

let slice before after s = narrow before (width_of s - after) s

let rec pp fmt = function
  | Empty -> Fmt.pf fmt "%s" "{}"
  | Singleton (w, z) -> Fmt.pf fmt "%a on %u bits" Z.pp_print z w
  | Wildcard w -> Fmt.pf fmt "%u-bit wide wildcard" w
  | Neg s -> Fmt.pf fmt "neg (%a)" pp s
  | Concat (s1, s2) -> Fmt.pf fmt "%a . %a" pp s1 pp s2
  | Union (s1, s2) -> Fmt.pf fmt "%a union %a" pp s1 pp s2
  | Inter (s1, s2) -> Fmt.pf fmt "%a inter %a" pp s1 pp s2
*)
