include Peahell.Make(struct

    let dot_output = ref false
    let smt_output = ref false
    let debug = Matching.debug
    let heuristics = Matching.enable_heuristics

    let default_repr = "caml"
    let available_reprs : (string * (module Repr.Full)) list = [
      "boxed", (module CompReprs.Boxed) ;
      "caml", (module CompReprs.Caml) ;
      (* "caml-noncomp", (module Repr.Camlike) ; *)
      (*"scala", (module CompReprs.Scala) ;*)
      "packed", (module PackedRepr) ;
      "niche", (module CompReprs.Niche) ;
    ]
    let repr = ref (List.assoc default_repr available_reprs)
    
    let name = "ribbit"
    type input = Syntax.program
    let options = [
      ("-g", Arg.Set dot_output, "Output graphs");
      ("--smt", Arg.Set smt_output, "Dump SMTLIB assertions");
      ("--debug", Arg.Set debug, "Debug matching compilation");
      ("-H", Arg.Bool ((:=) heuristics), "Use heuristics. Default to true");
      ("-r", Arg.Symbol
         (List.map fst available_reprs,
          (fun s -> repr := List.assoc s available_reprs)),
       "Choose the used repr"
      );
    ]

    type environment = Types.evaled_env * Types.env
    let initial_environment = Name.Map.empty, Name.Map.empty

    let read_more str =
      let i = ref (String.length str - 1) in
      while !i >= 0 && List.mem str.[!i] [' '; '\n'; '\t'; '\r'] do decr i done ;
      !i < 1 || (str.[!i] <> ';' || str.[!i - 1] <> ';')

    let file_parser =
      let f _name lexbuf =
        Peahell.Input.wrap (Parser.program Lexer.token) lexbuf
      in
      Some f
    let toplevel_parser =
      let f lexbuf =
        Peahell.Input.wrap (Parser.toplevel_phrase Lexer.token) lexbuf
      in
      Some f
    let expect_parser =
      let f _name lexbuf =
        Peahell.Input.wrap (Parser.expect_file Lexer.token) lexbuf
      in
      Some ( "(*EXPECT", "EXPECT*)", f)

    let exec fmt _import (ty_env, slot_env) l =
      let module Repr = (val !repr) in
      let module Matcher = Matching.Make(Repr) in

      let dot_files = ref [] in
      let xdot pp_dot stuff =
        let s = Filename.temp_file "ribbit_" ".dot" in
        let fmt = Format.formatter_of_out_channel @@ open_out s in
        Fmt.pf fmt "%a@." pp_dot stuff;
        dot_files := s :: !dot_files
      in
      
      let smt_file, smt_outchan = if !smt_output
        then let f = Filename.temp_file "ribbit_" ".smt" in
          f, Some (Format.formatter_of_out_channel @@ open_out f)
        else "", None in

      let dump_smt ty_sym ty_env unsolved_repr = match smt_outchan with
        | Some fmt ->
          Fmt.pf fmt "(*** Repr of type %s in %a ***)@.%a@."
            ty_sym Types.pp_evaled_env ty_env
            Mem.UnsolvedRepr.pp_smtlib (Mem.Hole.(MHole Id), unsolved_repr)
        | None -> () in

      let f (ty_env, slot_env) = function
        | Syntax.Declaration { name; parameters; definition } ->
          let ty_env = Simplify.declaration slot_env { name; parameters; definition } in
          let body = Name.Map.find name ty_env in
          Peahell.Report.printf "Declaration: %s = λ %a . %a@."
            name (Fmt.list Name.pp) parameters (Types.pp_evaled) body;
          let slot_env = Name.Map.add name (Types.Lam {parameters;body; body_env = ty_env}) slot_env in
          ty_env, slot_env
        | Show g ->
          let sym = Name.fresh () in
          let mono_env = Simplify.type_expr slot_env sym g in
          Peahell.Report.printf "Type: %s in %a@." sym Types.pp_evaled_env mono_env;
          let repr = (Repr.repr mono_env sym).generic in
          Peahell.Report.printf "Repr: %a@." Mem.SolvedRepr.pp repr;
          if !smt_output then
            dump_smt sym mono_env @@ Mem.ReprSkel.encode (Repr.pre_repr mono_env sym).generic;
          if !dot_output then xdot Mem.SolvedRepr.Dot.pp repr;
          Peahell.Report.printf "@.";
          ty_env, slot_env
        | Let (_n, ty, v) ->
          let sym = Name.fresh () in
          let mono_env = Simplify.type_expr slot_env sym ty in
          let vr = Repr.to_repr mono_env sym v in
          Peahell.Report.printf "Repr of %a:@.%a@." Value.pp v Mem.pp_value vr;
          (*if !smt_output then dump_smt @@ Repr.to_pre_repr mono_env sym v;*)
          if !dot_output then xdot Mem.Dot.pp vr;
          ty_env, slot_env
        | Matching m ->
          let dt = Matcher.compile_init (ty_env, slot_env) m in
          Peahell.Report.printf "Matching:@.%a@." Target.pp_dtree dt;
          if !dot_output then xdot Target.pp_dot dt;
          ty_env, slot_env
      in
      let res = List.fold_left f (ty_env, slot_env) l in
      List.iter
        (fun s -> let _ = Unix.system @@ Fmt.str "xdot %s" @@ Filename.quote s in ())
        (List.rev !dot_files);
      if !smt_output then
        Peahell.Report.printf "SMTLIB dumped to %s" @@ Filename.quote smt_file;
      res

  end)
