(** A hole denotes a position in a term *)
type t =
  | Hole | HRef of t
  | HTup of { arity: int; pos: int; hole: t }
  | HConstr of { constr: Name.t; arity: int; pos: int; hole: t }
type context = t

let rec subst h s = match h with
  | Hole -> s
  | HRef h' -> HRef (subst h' s)
  | HTup { arity; pos; hole } -> HTup { arity; pos; hole = subst hole s }
  | HConstr { constr; arity; pos; hole } ->
    HConstr { constr; arity; pos; hole = subst hole s }

let rec pp fmt h =
  let list_n_any n = List.init n (fun _ -> "_")
  and pp_list_commas = Fmt.list ~sep:Fmt.comma Fmt.string in
  match h with
  | Hole -> Fmt.pf fmt "□"
  | HRef h -> Fmt.pf fmt "&%a" pp h
  | HTup {arity; pos; hole} -> Fmt.pf fmt "<%a, %a, %a>"
      pp_list_commas (list_n_any pos)
      pp hole
      pp_list_commas (list_n_any (arity - pos - 1))
  | HConstr {constr; arity; pos; hole} -> Fmt.pf fmt "%s(%a, %a, %a)" constr
      pp_list_commas (list_n_any pos)
      pp hole
      pp_list_commas (list_n_any (arity - pos - 1))


module Head = struct
  type t =
    | Hole
    | Ref
    | Tuple of { arity: int; pos: int }
    | Constr of { constr: Name.t; arity: int; pos: int}

  let as_context : t -> context = function
    | Hole -> Hole
    | Ref -> HRef Hole
    | Tuple {arity; pos} -> HTup {arity; pos; hole = Hole}
    | Constr {constr; arity; pos} -> HConstr {constr; arity; pos; hole = Hole}

  let pp fmt h = pp fmt @@ as_context h
end

let by_induction (subst : 'a -> 'a -> 'a) (f : Types.evaled_env -> Name.t -> Head.t -> 'a)
= let rec aux (ty_env : Types.evaled_env) (sym0 : Name.t) (c : context) : 'a =
    let app_and_subst (h : Head.t) x = subst (f ty_env sym0 h) x in
    match Types.find_and_unroll sym0 ty_env, c with
    | _, Hole -> f ty_env sym0 Hole
    | Ref sym, HRef h ->
      app_and_subst Ref @@ aux ty_env sym h
    | Product syms, HTup { arity; pos; hole } ->
      let sym = List.nth syms pos in
      app_and_subst (Tuple { arity; pos }) @@ aux ty_env sym hole
    | Sum l, HConstr {constr; arity; pos; hole} ->
      let sym = List.nth (List.assoc constr l) pos in
      app_and_subst (Constr { constr; arity; pos }) @@ aux ty_env sym hole
    | _ -> invalid_arg "Hole.by_induction"
  in
  aux
