open Peahell
open Syntax
open Types
open Target

let debug = ref false
let enable_heuristics = ref true

(** Error reporting *)

type error =
  | Tag_not_in_type of Tag.t * Types.monomorph
  | Pattern_not_of_type of Syntax.pattern * Types.monomorph
  | Unexpected_pattern of Syntax.pattern
exception Error of error
let error err = raise @@ Error err
let prepare_error = function
  | Error (Tag_not_in_type (t, ty)) ->
    Some (Report.errorf
            "@[<v 2> The tag@;<1 2>%a@ doesn't belong to the type@;<1 2>%a@]"
            Tag.pp t pp_evaled ty)
  | Error (Pattern_not_of_type (pat, ty)) ->
    Some (Report.errorf
            "@[<v 2> The pattern@;<1 2>%a@ is not of type@;<1 2>%a@]"
            pp_pattern pat pp_evaled ty)
  | _ -> None
let () = Report.register_report_of_exn prepare_error

(** Utilities *)

let replace_nth n l new_ns =
  List.concat (List.mapi (fun k x -> if k = n then new_ns else [x]) l)

(** State handling *)

type environment =
  { ty_env : Types.evaled_env; discr : Name.t; discr_ty : Name.t }

module PatMat = struct

  type bindings = Target.rexpr Name.Map.t
  type line = {
    pats : Syntax.pattern list ;
    bindings : bindings ;
    body : Syntax.expr
  }
  (* header = matched subterm pos, subtype symbol *)
  type t = { headers : (Hole.t * Name.t) list; cases : line list }

  let line ({cases; _} : t) i = List.nth cases i
  let column ({cases; headers} : t) i =
    (List.nth headers i, List.map (fun l -> List.nth l.pats i) cases)
  let header ({headers; _} : t) i = List.nth headers i

  let get (m : t) ~line:l ~col:c = List.nth (line m l).pats c

  let map_line f ({cases; headers} : t) : t =
    { cases = List.map f cases; headers }

  let concat_map_line f ({cases; headers} : t) : t =
    { cases = List.concat_map f cases; headers }

  let expand_at patmatrix ~col f new_headers =
    let map_line { pats ; bindings ; body } =
      let sub_matrix = f (List.nth pats col) in
      let aux l =
        List.concat (List.mapi (fun k x -> if k = col then l else [x]) pats)
      in
      List.map (fun pats -> { pats = aux pats ; bindings ; body}) sub_matrix
    in
    { headers = replace_nth col patmatrix.headers new_headers;
      cases = CCList.concat_map map_line patmatrix.cases }

  let pp fmt (patmat : t) =
    let aux fmt {pats ; bindings ; body } =
      Fmt.pf fmt "@[<2>[ @[%a@] ->@ @[%a@] ]@]"
        (Fmt.list ~sep:Fmt.sp Syntax.pp_pattern) pats
        Target.pp_dtree (Success (bindings, body))
    in
    Fmt.pf fmt "@[%a@,%a@]"
      (Fmt.list ~sep:Fmt.cut (Fmt.pair Hole.pp Name.pp)) patmat.headers
      (Fmt.list ~sep:Fmt.cut aux) patmat.cases
end

let pp_env fmt {ty_env; discr; discr_ty} =
  Fmt.pf fmt "@[%s : %s@]" discr discr_ty

let get_variant_tuple ~constr l =
  match CCList.assoc_opt ~eq:Name.equal constr l with
  | Some t -> t
  | None ->
    error @@ Tag_not_in_type (Tag.Constr constr, Types.Sum l)

(** [get_tags env m i] returns all tags in column [m(i)] *)
let get_tags patmatrix : int -> Tag.Set.t =
  let rec aux tags p = match p with
    | PInt(z) -> Tag.Set.add (Tag.Int(z)) tags
    | PConstr(k) -> Tag.Set.add (Tag.Constr(k.constructor)) tags
    | POr(p1, p2) -> aux (aux tags p2) p1
    | PVar _ | PAny -> tags
    | PTuple _ | PRef _ -> tags
  in function i ->
    let _, c = PatMat.column patmatrix i in
    List.fold_left aux Tag.Set.empty c

let get_cases patmatrix ty_env ty_sym i =
  match Types.find_and_unroll ty_sym ty_env with
  | Sum l ->
    let tags = get_tags patmatrix i in
    let all_tags =
      Tag.Set.of_list @@ List.map (fun x -> Tag.Constr(fst x)) l
    in
    if Tag.Set.equal tags all_tags then
      tags
    else
      Tag.Set.add Top tags
  | Word _ ->
    let tags = get_tags patmatrix i in
    Tag.Set.add Top tags
  | Ref _
  | Product _ -> Tag.Set.singleton Top
  | Var x -> failwith x
  | _ -> .

module Heuristics = struct
  
  type t = ty_env:Types.evaled_env -> PatMat.t -> col:int -> int

  let select_best_columns (f : t) ty_env patmat columns =
    let rec aux current_max current_bests cols =
      match cols with
      | [] -> current_bests
      | col :: rest ->
        let score = f ~ty_env patmat ~col in
        if score > current_max then aux score [col] rest
        else if score = current_max then
          aux current_max (col :: current_bests) rest
        else aux current_max current_bests rest
    in
    List.rev @@ aux min_int [] columns
  
  let choose_with_heuristics ty_env heuristics patmat =
    let columns0 = CCList.( 0 --^ List.length (PatMat.line patmat 0).pats ) in
    let rec aux (heuristics : t list) columns = 
      match heuristics with
      | [] -> List.hd columns
      | f :: rest ->
        let columns = select_best_columns f ty_env patmat columns in
        match columns with
        | [x] -> x
        | _ -> aux rest columns
    in
    aux heuristics columns0
  
  let is_wildcard = function
    | Syntax.PAny -> true
    | _ -> false

  let first_row : t = fun ~ty_env patmat ~col ->
    if is_wildcard @@ PatMat.get patmat ~col ~line:0 then
      0
    else
      1

  let small_default : t = fun ~ty_env patmat ~col ->
    let (_, c) = PatMat.column patmat col in
    -(CCList.count is_wildcard c)

  let small_branching_factor : t = fun ~ty_env patmat ~col ->
    let (_, ty) = PatMat.header patmat col in
    let tags = get_cases patmat ty_env ty col in
    -(Tag.Set.cardinal tags)

  let arity : t = fun ~ty_env patmat ~col ->
    let (_, ty_sym) = PatMat.header patmat col in
    let a = 
      let ty = Types.find_and_unroll ty_sym ty_env in
      match ty with
      | Sum l ->
        let tags = get_tags patmat col in
        let get_constr_arity tag =
          let constr = match tag with Tag.Constr c -> c | _ -> assert false in
          List.length @@ get_variant_tuple ~constr l
        in
        Tag.Set.fold (fun tag acc -> get_constr_arity tag + acc) tags 0
      | Word _ -> 1
      | Ref _ -> 1
      | Product l -> List.length l
      | _ -> assert false
    in
    -a

  (** [count_prefix p l] counts the length of the prefix of [l] such that [p] is
      true on all elements in the prefix.
  *)
  let rec count_prefix p l =
    List.length @@ CCList.take_while p l
  
  let constructor_prefix : t = fun ~ty_env patmat ~col ->
    let (_, c) = PatMat.column patmat col in
    let rec is_generalized_constructor = function
      | PConstr _
      | PInt _
      | PRef _
      | PTuple _
        -> true
      | POr (p, _) -> is_generalized_constructor p
      | PAny | PVar _ -> false
    in
    count_prefix is_generalized_constructor c

  let chosen_heuristics = [ constructor_prefix ; small_branching_factor ; arity ]

  let choose env patmat =
    let hs =
      if !enable_heuristics then chosen_heuristics else []
    in
    choose_with_heuristics env.ty_env hs patmat
end


module Make (R:Repr.S) = struct

  (** Introduce let-bindings based on variables. Use {!R.knit} *)
  let introduce_bindings env patmat =
    let do_one_col patmat i =
      let (hi, tyi) = PatMat.header patmat i in
      let aux p bindings = match p with
        | PVar(x) ->
          PAny, Name.Map.add x (R.knit env.ty_env env.discr_ty hi env.discr) bindings
        | POr(p1, p2) ->
          error @@ Unexpected_pattern p (*aux p1 @@ aux p2 @@ bindings*)
        | _ -> p, bindings
      in
      PatMat.map_line 
        (fun {pats; bindings; body} -> let p, b = aux (List.nth pats i) bindings in
           { pats = replace_nth i pats [p]; bindings = b; body })
        patmat
    in List.fold_left do_one_col patmat (List.mapi (fun i _ -> i) patmat.headers)

  let split_or_pats env patmat =
    let rec do_one_pat = function
      | POr(p1, p2) -> (do_one_pat p1) @ (do_one_pat p2)
      | p -> [[p]]
    in
    let do_one_line col {PatMat. pats; bindings; body} =
      let split_pats = do_one_pat (List.nth pats col) in
      let new_pats = List.map (replace_nth col pats) split_pats in
      List.map (fun ps -> {PatMat. pats = ps; bindings; body }) new_pats
    in
    let do_one_col patmat col = PatMat.concat_map_line (do_one_line col) patmat in
    List.fold_left do_one_col patmat (List.mapi (fun i _ -> i) patmat.headers)

  let expand env patmatrix ~(col : int) (t : Tag.t) : PatMat.t =
    let (hi, tyi) = PatMat.header patmatrix col in
    let tyi = find_and_unroll tyi env.ty_env in
    let new_headers =
      match tyi, t with
      | (Word _, _) | (Sum _, Top) -> []
      | Ref rty, Top -> [Hole.HRef Hole, rty]
      | Product l, Top ->
        let arity = List.length l in
        List.mapi (fun pos ty -> Hole.HTup {arity; pos; hole=Hole}, ty) l
      | Sum l, Tag.Constr constr ->
        let variant_tuple = get_variant_tuple ~constr l in
        let arity = List.length variant_tuple in
        List.mapi (fun pos ty -> Hole.HConstr {constr; arity; pos; hole=Hole}, ty)
          variant_tuple
      | ty, tag -> error @@ Tag_not_in_type (tag, ty)
    in let new_headers =
      (List.map (fun (h,ty) -> Hole.subst hi h, ty) new_headers) in
    let mk_pat = match tyi, t with
      | Word _, Tag.Int c -> begin function
          | PInt c' -> if c' = c then [[]] else []
          | PAny -> [[]]
          | pat -> error @@ Pattern_not_of_type (pat, tyi)
        end
      | Word _, Top -> begin function
          | PInt _ -> []
          | PAny -> [[]]
          | pat -> error @@ Pattern_not_of_type (pat, tyi)
        end
      | Ref _, Top -> begin function
          | PRef p -> [[p]]
          | PAny -> [[PAny]]
          | pat -> error @@ Pattern_not_of_type (pat, tyi)
        end
      | Product l, Top ->
        let n = List.length l in begin function
          | PTuple l' when List.length l' = n -> [l']
          | PAny -> [List.init n (fun _ -> PAny)]
          | pat -> error @@ Pattern_not_of_type (pat, tyi)
        end
      | Sum l, Tag.Constr constr ->
        let variant_tuple = get_variant_tuple ~constr l in
        let n = List.length variant_tuple in
        begin function
          | PConstr {constructor; arguments} when constructor = constr ->
            [arguments]
          | PConstr _ -> []
          | PAny -> [List.init n (fun _ -> PAny)]
          | pat -> error @@ Pattern_not_of_type (pat, tyi)
        end
      | Sum l, Top -> begin function
          | PConstr _ -> []
          | PAny -> [[]]
          | pat -> error @@ Pattern_not_of_type (pat, tyi)
        end 
      | _ -> error @@ Tag_not_in_type (t, tyi)
    in
    PatMat.expand_at patmatrix ~col mk_pat new_headers

  let rec compile_branch env pats : int -> Tag.t -> dtree = fun i t ->
    compile env (expand env pats ~col:i t)

  and compile env patmat : dtree =
    let rec only_any p = match p with
      | PAny -> true
      | PTuple l -> List.for_all only_any l
      | POr(p1, p2) -> only_any p1
      | _ -> false
    in
    if !debug then 
      Report.eprintf
        "@[<v 2>Compiling:@ @[<v 2>environment @,%a@]@ @[<v 2>patterns @,%a@]@."
        pp_env env PatMat.pp patmat;
    let patmat = introduce_bindings env (split_or_pats env patmat) in
      if !debug then
        Report.eprintf "With split or-pats and lifted vars: %a@." PatMat.pp patmat;
    match patmat.cases with
    | [] -> Fail
    | { pats ; bindings ; body } :: _ when List.for_all only_any pats ->
      Success (bindings, body)
    | _ ->
      let i = Heuristics.choose env patmat in
      let hi, tyi = PatMat.header patmat i in
      if !debug then
        Report.eprintf
          "@[<v2>Picked column %i:@,@[<2>- %a@]@,@[<2>- %a@]@]@."
          i Hole.pp hi Types.pp_evaled (Types.find_and_unroll tyi env.ty_env);
      let subtrees =
        get_cases patmat env.ty_env tyi i
        |> Tag.Set.to_iter
        |> Iter.map (fun t -> (t, compile_branch env patmat i t))
        |> Tag.Map.of_iter
      in
      if !debug then
        Report.eprintf
          "Frogging: %a@."
          Fmt.(list ~sep:semi (pair ~sep:(any " -> ") Tag.pp Target.pp_dtree))
          (Tag.Map.bindings subtrees);
      R.frog env.ty_env env.discr_ty hi subtrees env.discr

  let compile_init (ty_env, slot_env) m =
    let discr, discr_ty = Name.unique_discr in
    let ty_env = Simplify.type_expr slot_env discr_ty m.pattern_type in
    let headers = [ Hole.Hole, discr_ty ]
    and cases =
      let mk_line {pattern;expr} =
        {PatMat. pats = [pattern] ; bindings = Name.Map.empty ; body = expr }
      in
      List.map mk_line m.clauses
    in
    Dtree.share @@ compile { ty_env; discr; discr_ty } {PatMat. cases; headers}
end
