(** Formalization of memory values *)

(** A value in memory *)
type immediate =
  | Symbol of Name.t
  | Word of Z.t
  | True
  | False
let pp_immediate fmt = function
  | Symbol x -> Fmt.pf fmt "%s" x
  | Word z -> Fmt.pf fmt "%a" Z.pp_print z
  | True -> Fmt.pf fmt "True"
  | False -> Fmt.pf fmt "False"
let as_z = function
  | Symbol x -> Z.zero (* TODO *)
  | Word z -> z
  | True | False -> Z.one
let smt_expr_of_imm = function
  | Symbol x -> SMT.(expr_of_zero @@ bv_sort 1) (* TODO *)
  | Word z ->
    let len = if z = Z.zero then 1 else Z.numbits z in
    SMT.(expr_of_z (bv_sort len) z)
  | True -> SMT.bool_true
  | False -> SMT.bool_false
let imm_of_smt_expr =
  SMT.eval_expr (function false -> False | true -> True) (fun z -> Word z)
type t =
  | VImm of { width: int; bits: immediate }
  | VPtr of { width: int; align: int; pointee: t; bits: immediate }
  | VBlock of t list
type value = t
let rec pp_value fmt = function
  | VImm { width; bits } ->
    Fmt.pf fmt "(%a)%i" pp_immediate bits width
  | VPtr { width; align; pointee; bits } ->
    Fmt.pf fmt "&%i,%i(%a;%a)" width align pp_value pointee pp_immediate bits
  | VBlock l ->
    Fmt.pf fmt "{ %a }" (Fmt.list ~sep:Fmt.semi pp_value) l

module SrcHole = Hole
(** Memory contexts with embedded word op expressions *)
module Hole = struct
  module WordOps = struct
    type t =
      | Id
      | Cmp of t * Target.rcmp * Z.t
      | Arith of t * Target.rarith * Z.t
      | Extract of { ops : t; low : int; high : int }

    let rec subst ops ops' = match ops with
      | Id -> ops'
      | Cmp (ops, cmp, z) -> Cmp (subst ops ops', cmp, z)
      | Arith (ops, ar, z) -> Arith (subst ops ops', ar, z)
      | Extract { ops; low; high } -> Extract { ops = subst ops ops'; low; high }
    
    (* max val of ops^-1(res) *)
    let rec max_input_val bv_max ops res = match ops with
      | Id -> as_z res
      | Cmp (ops, cmp, z) -> (match cmp, res with
        | (Eq | Leq), True | (Neq | Gt), False ->
          max_input_val bv_max ops (Word z)
        | Lt, True | Geq, False ->
          max_input_val bv_max ops (Word Z.(z - one))
        | _ -> bv_max)
      | Arith (ops, ar, z) -> (match ar, res with
        | Mod, _ -> bv_max
        | _, Word z' -> max_input_val bv_max ops (Word (Target.arith_to_z ar z' z))
        | _ -> bv_max)
      | Extract { ops; low; high } -> bv_max

    let rec pp id_sym fmt = function
      | Id -> Fmt.pf fmt "%s" id_sym
      | Cmp (ops, cmp, z) ->
        Fmt.pf fmt "%a %s %a" (pp id_sym) ops (Target.cmp_to_str cmp) Z.pp_print z
      | Arith (ops, ar, z) ->
        Fmt.pf fmt "(%a %s %a)" (pp id_sym) ops (Target.arith_to_str ar) Z.pp_print z
      | Extract { ops; low; high } ->
        Fmt.pf fmt "extract(%i, %i)(%a)" low high (pp id_sym) ops

    let rec apply ops imm = match ops, imm with
      | Id, _ -> imm
      | _ ->
        let sym = Fmt.str "%a" pp_immediate imm in
        Symbol (Fmt.str "%a" (pp sym) ops)

    let rec apply_to_smt_expr ops (expr : SMT.sorted_expr) =
      let open SMT in match ops with
      | Id -> expr
      | Cmp (ops, cmp, z) ->
        let { expr = lhs; sort } = apply_to_smt_expr ops expr in
        let rhs = expr_of_z sort z
        and smt_cmp = (match cmp with
          | Eq -> eq_expr | Neq -> neq_expr
          | Leq -> bv_ule | Lt -> bv_ult
          | Geq -> bv_uge | Gt -> bv_ugt) in
        { sort = bool_sort; expr = smt_cmp lhs rhs }
      | Arith (ops, ar, z) ->
        let { expr = lhs; sort } = apply_to_smt_expr ops expr in
        let rhs = expr_of_z sort z
        and smt_ar = (match ar with
          | Add -> bv_add | Sub -> bv_sub
          | Mul -> bv_mul | Div -> bv_div | Mod -> bv_mod) in
        { sort = expr.sort; expr = smt_ar lhs rhs }
      | Extract { ops; low; high } ->
        let { expr; _ } = apply_to_smt_expr ops expr in
        { sort = bv_sort (high - low + 1);
          expr = bv_extract high low expr }

    let rec focus ops e = match ops with
      | Id -> e
      | Cmp (ops, cmp, z) ->
        Target.RCmp (focus ops e, cmp, z)
      | Arith (ops, ar, z) ->
        Target.RArith (focus ops e, ar, z)
      | Extract { ops; low; high } ->
        let cutoff = high + 1 in
        let mask = Z.((one + one) ** cutoff - one) in
        let e = focus (ops) e in
        Target.(RRShift (RAnd (e, mask), low))

  end

  type t =
    | MHole of WordOps.t
    | MHPtr of { len: int; align: int; hole: t }
    | MHBlock of { pos: int; hole: t }
  
  let max_input_val bv_max pos imm = match pos with
    | MHole ops -> WordOps.max_input_val bv_max ops imm
    | _ -> assert false

  let rec pp fmt = function
    | MHole Id -> Fmt.pf fmt "■"
    | MHole ops -> Fmt.pf fmt "■(%a)" (WordOps.pp "y") ops
    | MHPtr { len; align; hole } -> Fmt.pf fmt "&(%d/%d)(%a)" len align pp hole
    | MHBlock { pos; hole } -> Fmt.pf fmt "{{ ...; %a (%d); ... }}" pp hole pos

  let rec subst mh mh' = match mh with
    | MHole Id -> mh'
    | MHole ops -> (match mh' with
      | MHole ops' -> MHole (WordOps.subst ops' ops) (* ops, ops' are swapped on purpose *)
      (* TODO deal with pointers too? *)
      | _ -> assert false)
    | MHPtr { len; align; hole } ->
      MHPtr { len; align; hole = subst hole mh' }
    | MHBlock { pos; hole } ->
      MHBlock { pos; hole = subst hole mh' }

  let rec focus mh e = match mh with
    | MHole ops -> WordOps.focus ops e
    | MHPtr { len; hole } -> focus hole (Target.RDeref e)
    | MHBlock { pos; hole } -> focus hole (Target.RAccess (e, pos, 64))
  let focus mh e = Target.simplify_rexpr @@ focus mh e

  let rec is_prefix_of mh mh' = match mh, mh' with
    | MHole Id, MHole Id -> false
    | MHole Id, _ -> true
    | MHole (Arith (ops, ar, z)), MHole (Arith (ops', ar', z')) ->
      z = z' && ar = ar' && is_prefix_of (MHole ops) (MHole ops')
    | MHole (Cmp (ops, cmp, z)), MHole (Cmp (ops', cmp', z')) ->
      z = z' && cmp = cmp' && is_prefix_of (MHole ops) (MHole ops')
    | MHole (Extract { ops; low; high }),
      MHole (Extract { ops = ops'; low = low'; high = high' }) ->
      low = low' && high = high' && is_prefix_of (MHole ops) (MHole ops')
    | MHole (Extract { high; _ }), MHPtr { align; _ } when high < align ->
      true (* check pointer-ness BEFORE dereferencing *)
    | MHPtr { len; align; hole }, MHPtr { len = len'; align = align'; hole = hole' } ->
      len = len' && align = align' && is_prefix_of hole hole'
    | MHBlock { pos; hole }, MHBlock { pos = pos'; hole = hole' } ->
      pos = pos' && is_prefix_of hole hole'
    | _ -> false

  let rec non_prefix_compat mh mh' = match mh, mh' with
    | MHole _, MHole _ -> not (is_prefix_of mh mh' || is_prefix_of mh' mh)
    | MHole _, MHPtr _ | MHPtr _, MHole _ ->
      not (is_prefix_of mh mh' || is_prefix_of mh' mh)
    | MHPtr { len; align; hole }, MHPtr { len = len'; align = align'; hole = hole' } ->
      len = len' && align = align' && non_prefix_compat hole hole'
    | MHBlock { pos; hole }, MHBlock { pos = pos'; hole = hole' } ->
      pos <> pos' || non_prefix_compat hole hole'
    | _ -> false

  let cmp mh mh' =
    if mh = mh' || non_prefix_compat mh mh' then 0
    else if is_prefix_of mh mh' then -1
    else if is_prefix_of mh' mh then 1
    else failwith "coin coin"

end

(** All-purpose memory patterns parametrised by word specs *)
module type WordPat = sig
  type t
  val pp : Stdlib.Format.formatter -> t -> unit
  val neutral : t
  val apply_wordops : Hole.WordOps.t -> t -> t
  val inter : t -> t -> t
  val union : t -> t -> t
end
module Pattern (W:WordPat) = struct
  type t =
    | Empty
    | App of { ty_sym: Name.t; term_pos: SrcHole.t }
    | Word of { len: int; max_val: Z.t; bits: W.t }
    | Ptr of { len: int; align: int; mpat: t; bits: W.t }
    | Block of (t option) list
    | Union of t list

  let walk_blocks f l1 l2 =
    let rec aux l1 l2 acc = match l1, l2 with
      | l, [] | [], l -> List.rev_append acc l
      | None::t1, None::t2 -> aux t1 t2 (None :: acc)
      | None::t1, (Some h)::t2 | (Some h)::t1, None::t2 ->
        aux t1 t2 ((Some h) :: acc)
      | (Some h1)::t1, (Some h2)::t2 ->
        aux t1 t2 ((Some (f h1 h2)) :: acc)
    in
    aux l1 l2 []

  let rec pp fmt = function
    | Empty -> Fmt.pf fmt "_"
    | App { ty_sym; term_pos } ->
      Fmt.pf fmt "repr_(%s)(subterm at %a)" ty_sym SrcHole.pp term_pos
    | Word { len; max_val; bits } ->
      Fmt.pf fmt "Word%i,≤%a(%a)" len Z.pp_print max_val W.pp bits
    | Ptr { len; align; mpat; bits } ->
      Fmt.pf fmt "&%i,%i(%a)(%a)" len align pp mpat W.pp bits
    | Block l -> 
      Fmt.pf fmt "{ %a }" (Fmt.list ~sep:Fmt.semi (Fmt.option pp)) l
    | Union l -> 
      Fmt.pf fmt "%a" (Fmt.list ~sep:(Fmt.any " ⊔ ") pp) l

  (** Dot output *)
  module Dot = struct
    type id = int
    type blob =
      | DEmpty
      | DWord of { width: int; max_val: Z.t; bits: W.t }
      | DPtr of { width: int; align: int; target: id; bits: W.t }
      | DApp of { fsym: Name.t; term_pos: SrcHole.t }
      | DUnion of id list
    type contents =
      | DBlob of blob
      | DBlock of contents list
    type vertex = { id: id; contents: contents }
    type edge =
      | EPointer of id * id
      | EUnion of id * id
      
    let as_dot_graph t =
      let r = ref 0 in
      let new_id () = incr r; !r in
      let vertices : vertex list ref = ref []
      and edges : edge list ref = ref [] in
      let add_vertex x = vertices := x :: !vertices
      and add_edge x = edges := x :: !edges in
      let rec gather_vertices mpat =
        let id = new_id () in
        add_vertex { id; contents = contents_of_mpat id mpat };
        id
      and contents_of_mpat vertex_id = function
        | Empty -> DBlob DEmpty
        | App { ty_sym; term_pos } ->
          DBlob (DApp { fsym = ty_sym; term_pos })
        | Word { len; max_val; bits } ->
          DBlob (DWord { width = len; max_val; bits })
        | Ptr { len; align; mpat; bits } ->
          let target = gather_vertices mpat in
          add_edge (EPointer (vertex_id, target));
          DBlob (DPtr { width = len; align; target; bits })
        | Block l ->
          DBlock (List.map
            (function
            | None -> DBlob DEmpty
            | Some mp -> contents_of_mpat vertex_id mp)
            l)
        | Union l ->
          let branches = List.map (fun mpat ->
              let target = gather_vertices mpat in
              add_edge (EUnion (vertex_id, target));
              target)
            l in
          DBlob (DUnion branches)
      in
      let _ = gather_vertices t in
      !vertices, !edges

    let pp_blob ppf = function
      | DEmpty -> Fmt.pf ppf "∅"
      | DWord { width; max_val; bits } ->
        Fmt.pf ppf "%a (≤ %a on %i bits)" W.pp bits Z.pp_print max_val width
      | DPtr { width; align; bits; _ } ->
        Fmt.pf ppf "&(%i/%i) %a" width align W.pp bits
      | DApp { fsym; term_pos } ->
        Fmt.pf ppf "%s-repr of the subterm at %a" fsym SrcHole.pp term_pos
      | DUnion _ -> Fmt.pf ppf "∐"

    let pp_vertex ppf { id; contents } =
      let attrs =
        let rec pp_label inside_block ppf = function
          | DBlob blob ->
            if inside_block
            then Fmt.pf ppf "<%i> %a" id pp_blob blob
            else Fmt.pf ppf "%a" pp_blob blob
          | DBlock contents ->
            let nest = if inside_block
              then Fmt.pf ppf "{ nested block | { %a } }"
              else Fmt.pf ppf "%a" in
            nest
              (Fmt.list ~sep:(Fmt.any " | ") (pp_label true))
              contents
        and shape = match contents with
          | DBlock _ -> "record"
          | DBlob (DUnion _) -> "circle"
          | _ -> "box"
        in
        [ ("label", Fmt.str "%a" (pp_label false) contents);
          ("shape", shape) ]
      in
      let pp_attr ppf (field, v) = Fmt.pf ppf {|%s="%s"|} field v in
      Fmt.pf ppf "%d [%a];" id (Fmt.list ~sep:(Fmt.any ", ") pp_attr) attrs

    let pp_edge ppf e =
      let attrs, id, id' = match e with
        | EPointer (id, id') -> [], id, id'
        | EUnion (id, id') ->
          ["color", "forestgreen"; "arrowhead", "none"], id, id'
      in
      let pp_attr ppf (field, v) = Fmt.pf ppf "%s=%S" field v in
      Fmt.pf ppf "%d -> %d:n [%a];"
        id id'
        (Fmt.list ~sep:(Fmt.any ", ") pp_attr)
        attrs

    let pp ppf t =
      let vertices, edges = as_dot_graph t in
      Fmt.pf ppf {|@[<v2>digraph G {@,concentrate=true;rankdir="LR";@,%a@,@,%a@,}@]|}
        (Fmt.list ~sep:Fmt.cut pp_vertex)
        vertices
        (Fmt.list ~sep:Fmt.cut pp_edge)
        edges
  end

  exception Notersect of t * t
  let rec inter (env : t Name.Map.t) mp1 mp2 = match mp1, mp2 with
    | Empty, mp | mp, Empty -> mp
    | App { ty_sym = t1; term_pos = h1 }, App { ty_sym = t2; term_pos = h2 }
      when t1 = t2 && h1 = h2 -> App { ty_sym = t1; term_pos = h1 }
    | App { ty_sym; term_pos }, mp | mp, App { ty_sym; term_pos } ->
      (match Name.Map.find_opt ty_sym env with
      | Some mp' when inter env mp mp' = mp' -> App { ty_sym; term_pos }
      | _ -> assert false)
    | Word { len = l1; max_val = mv1; bits = b1 },
      Word { len = l2; max_val = mv2; bits = b2 } ->
      Word { len = (if l1 >= l2 then l1 else l2);
             max_val = Z.max mv1 mv2;
             bits = W.inter b1 b2 }
    | Ptr { len; align; mpat = mp1; bits = b1 },
      Ptr { len = len'; align = align'; mpat = mp2; bits = b2 }
      when (len = len') && (align = align') ->
      Ptr { len; align; mpat = inter env mp1 mp2; bits = W.inter b1 b2 }
    | Word { len = wlen; max_val; bits = wbits }, Ptr { len; align; mpat; bits }
    | Ptr { len; align; mpat; bits }, Word { len = wlen; max_val; bits = wbits }
      when Z.numbits max_val <= align ->
      Ptr { len; align; mpat; bits = W.inter bits wbits }
    | Block l1, Block l2 ->
      Block (walk_blocks (inter env) l1 l2)
    | mp, Union l | Union l, mp -> Union (List.map (inter env mp) l)
    | _ -> raise (Notersect (mp1, mp2))

  let rec union env mp1 mp2 = match mp1, mp2 with
    | Empty, mp | mp, Empty -> mp
    | App { ty_sym = t1; term_pos = h1 }, App { ty_sym = t2; term_pos = h2 }
      when t1 = t2 && h1 = h2 -> App { ty_sym = t1; term_pos = h1 }
    | App { ty_sym; term_pos }, mp | mp, App { ty_sym; term_pos } ->
      (match Name.Map.find_opt ty_sym env with
      | Some mp' when union env mp mp' = mp' -> App { ty_sym; term_pos }
      | _ -> Union [mp1; mp2])
    | Word { len; max_val; bits },
      Word { len = len'; max_val = max_val'; bits = bits' } ->
      let len = (if len >= len' then len else len')
      and max_val = Z.max max_val max_val' in
      Word { len; max_val; bits = W.union bits bits' }
    | Ptr { len; align; mpat; bits },
      Ptr { len = len'; align = align'; mpat = mpat'; bits = bits' }
      when len = len' && align = align' ->
      Ptr { len; align; mpat = union env mpat mpat'; bits = W.union bits bits' }
    | Block l1, Block l2 ->
      Block (walk_blocks (union env) l1 l2)
    | Union l, mp | mp, Union l -> Union (mp :: l)
    | _ -> Union [mp1; mp2]

  let inter_list env = List.fold_left (inter env) Empty
  let union_list env = List.fold_left (union env) Empty

  let rec subst_in_hole (mh : Hole.t) mp = match mh with
    | MHole Id -> mp
    | MHole ops -> (match mp with
      | Empty -> Empty
      | Word { len; max_val; bits } ->
        let max_val = Hole.WordOps.max_input_val Z.(~$2 ** len - one) ops (Word max_val) in
        Word { len; max_val; bits = W.apply_wordops ops bits }
      | Ptr { len; align; mpat; bits } ->
        Ptr { len; align; mpat; bits = W.apply_wordops ops bits }
      | Union l ->
        Union (List.map (subst_in_hole mh) l)
      | _ -> assert false
      )
    | MHPtr { len; align; hole } ->
      Ptr { len; align; mpat = subst_in_hole hole mp; bits = W.neutral }
    | MHBlock { pos; hole } ->
      let mpats = List.init (pos + 1)
          (fun i -> if i = pos then Some (subst_in_hole hole mp) else None)
      in Block mpats
end

(** Vaguely LLVM-IR-like memory types *)
module WordTy : WordPat = struct
  type t = unit
  let pp fmt x = Fmt.pf fmt "%a" Fmt.nop x
  let neutral = ()
  let apply_wordops e t = ()
  let inter t1 t2 = ()
  let union t1 t2 = ()
end
module Type = Pattern(WordTy)
type typ = Type.t

(** Solved reprs with fully interpreted functions *)
module SolvedWord = struct
  type t = SMT.solution
  let neutral = [], SMT.{ cases = []; default = expr_of_zero (bv_sort 1) }
  let pp fmt (_, interp) =
    Fmt.pf fmt "%a" SMT.pp_interp interp

  let apply_wordops ops (params, SMT.{ cases; default }) =
    params,
    SMT.{ cases = List.map
            (fun (args, res) ->
              let res = Hole.WordOps.apply_to_smt_expr ops
                { expr = res; sort = sort_of res } in
              args, res.expr) 
            cases;
          default = (Hole.WordOps.apply_to_smt_expr ops
            { expr = default; sort = sort_of default }).expr }

  let inter imm1 imm2 =
    if imm1 = imm2 then imm1 else assert false
  let union imm1 imm2 =
    if imm1 = imm2 then imm1 else assert false
end
module SolvedRepr = struct
  include Pattern(SolvedWord)

  let rec to_memval (subty : SrcHole.t -> Name.t) (choose_pat: SrcHole.t -> 'a)
  (env: ('a -> t) Name.Map.t) : t -> value =
    let param_to_solution h =
      (match Name.Map.find (subty h) env @@ choose_pat h with
      | Word { bits; _ } -> bits
      | _ -> assert false)
    in function
    | App { ty_sym; term_pos } ->
      let mp = (Name.Map.find ty_sym env) (choose_pat term_pos) in
      to_memval subty choose_pat env mp
    | Word { len; max_val; bits } ->
      let e = SMT.eval_solution param_to_solution bits in
      let bits = imm_of_smt_expr e in
      VImm { width = len; bits }
    | Ptr { len; align; mpat; bits } ->
      let e = SMT.eval_solution param_to_solution bits in
      let bits = imm_of_smt_expr e in
      VPtr { width = len; align; pointee = to_memval subty choose_pat env mpat; bits }
    | Block l ->
      VBlock (List.map (function
          | None -> to_memval subty choose_pat env Empty
          | Some mp -> to_memval subty choose_pat env mp)
        l)
    | _ -> VImm { width = -1; bits = Symbol "pouet" }
  
end

module UnsolvedWord = struct
  type t = SMT.params * SMT.problem
  let neutral =
    [], SMT.{ declaration = Cst (const_expr (bv_sort 1) "zut"); assertion = bool_true }
  let apply_wordops = fun _ -> assert false
  let inter = fun _ -> assert false
  let union = fun _ -> assert false
  let pp fmt (params, pb) =
    Fmt.pf fmt "(%a) -> %a"
      Fmt.(list ~sep:semi @@ pair SrcHole.pp nop) params
      SMT.pp_problem pb
  let pp_smtlib fmt (_, pb) = SMT.pp_problem fmt pb
end
module UnsolvedRepr = struct
  include Pattern(UnsolvedWord)
  let solve ~concrete : t -> SolvedRepr.t =
    let solve_word (params, pb) : SolvedWord.t =
      params, SMT.solve_problem ~concrete pb in
    let rec aux : t -> SolvedRepr.t = function
      | Empty -> Empty
      | App { ty_sym; term_pos } -> App { ty_sym; term_pos }
      | Word { len; max_val; bits } ->
        Word { len; max_val; bits = solve_word bits }
      | Ptr { len; align; mpat; bits } ->
        Ptr { len; align; mpat = aux mpat; bits = solve_word bits }
      | Block l ->
        Block (List.map (function
            | None -> None
            | Some mp -> Some (aux mp))
          l)
      | Union l ->
        Union (List.map aux l)
    in aux

  let rec pp_smtlib fmt (cur_mpos, mpat) = match mpat with
    | Word { len; max_val; bits } ->
      Fmt.pf fmt "(* Word%i,≤%a at %a *)@.%a"
        len Z.pp_print max_val Hole.pp cur_mpos
        UnsolvedWord.pp_smtlib bits
    | Ptr { len; align; mpat; bits } ->
      Fmt.pf fmt "(* &%i,%i at %a *)@.%a@.%a"
        len align Hole.pp cur_mpos
        UnsolvedWord.pp_smtlib bits
        pp_smtlib (Hole.(subst cur_mpos @@ MHPtr { len; align; hole = MHole Id }), mpat)
    | Block l ->
      let fields = List.mapi (fun pos x ->
          Hole.(subst cur_mpos @@ MHBlock { pos; hole = MHole Id }), x)
        (List.filter_map (fun x -> x) l) in
      Fmt.pf fmt "(** Block at %a **)@.%a"
        Hole.pp cur_mpos
        Fmt.(list ~sep:cut pp_smtlib) fields
    | Union l ->
      let l = List.map (fun x -> cur_mpos, x) l in
      Fmt.pf fmt "(** Union at %a **)@.%a"
        Hole.pp cur_mpos
        Fmt.(list ~sep:cut pp_smtlib) l
    | _ -> ()

end

(** Repr skels with pre-SMT word formulae *)
module WordSkel = struct
  type rhs =
    | Imm of immediate
    | Var of SrcHole.t * int
  type branch =
    | Empty
    | Atom of Hole.WordOps.t * rhs
    | And of branch * branch
  type t = branch list
  let neutral = [ Empty ]

  let apply_wordops ops =
    let rec aux = function
      | Empty -> Empty
      | Atom (ops', r) ->
        Atom (Hole.WordOps.subst ops ops', r)
      | And (phi1, phi2) ->
        And (aux phi1, aux phi2)
    in List.map aux

  let inter l1 l2 =
    List.concat_map (fun phi1 -> List.map (fun phi2 -> And (phi1, phi2)) l2) l1
  let union = List.append

  let rec pp_branch fmt =
    let pp_wordops = Hole.WordOps.pp "y" in
    function
    | Empty -> Fmt.pf fmt "⊤"
    | Atom (ops, Imm imm) -> (match imm with
      | True -> Fmt.pf fmt "%a" pp_wordops ops
      | False -> Fmt.pf fmt "¬(%a)" pp_wordops ops
      | _ -> Fmt.pf fmt "%a = %a" pp_wordops ops pp_immediate imm)
    | Atom (ops, Var (h, width)) ->
      Fmt.pf fmt "%a = x_(%a:%d)" pp_wordops ops SrcHole.pp h width
    | And (phi1, phi2) ->
      Fmt.pf fmt "(%a) ∧ (%a)" pp_branch phi1 pp_branch phi2
  let pp fmt =
    Fmt.pf fmt "%a" Fmt.(list ~sep:(any " ∨ ") pp_branch)

  let rec simplify_branch = function
    | And (phi1, phi2) ->
      (match simplify_branch phi1, simplify_branch phi2 with
      | Empty, phi | phi, Empty -> phi
      | phi1, phi2 -> And (phi1, phi2))
    | phi -> phi
  let rec simplify = function
    | [] -> []
    | phi :: l -> (match simplify_branch phi with
      | Empty -> simplify l
      | phi -> phi :: simplify l)
  let rec as_immediate l =
    Symbol (Fmt.str "%a" pp (simplify l))

  (* SMT/Z3 stuff *)
  let imm_to_z3 bv_sort = function
    | Symbol x -> SMT.const_expr bv_sort x
    | Word z -> SMT.expr_of_z bv_sort z
    | True -> SMT.bool_true
    | False -> SMT.bool_false
  let rec ops_to_z3 (out_expr : SMT.sorted_expr) : Hole.WordOps.t -> SMT.sorted_expr =
    fun ops -> Hole.WordOps.apply_to_smt_expr ops out_expr

  let rec to_z3 params out_expr : branch -> SMT.sorted_expr =
    let open SMT in function
    | Empty -> { sort = bool_sort; expr = bool_true }
    | Atom (ops, Imm imm) ->
      (*Peahell.Report.fprintf Fmt.stderr "Atom: %a = %a@," (Hole.WordOps.pp "y") ops pp_immediate imm;*)
      let { sort; expr = lexp } = ops_to_z3 out_expr ops in
      let rexp = imm_to_z3 sort imm in
      (*Peahell.Report.fprintf Fmt.stderr "lhs: %s; rhs: %s@." (Z3.Expr.to_string lexp) (Z3.Expr.to_string rexp);*)
      { sort = bool_sort; expr = eq_expr lexp rexp }
    | Atom (ops, Var (h, len)) ->
      (*Peahell.Report.fprintf Fmt.stderr "Atom: %a@," pp_branch (Atom (ops, Var (h, len)));*)
      let { expr = lexp; _ } = ops_to_z3 out_expr ops in
      let { expr = rexp; sort } = List.assoc h params in
      (*Peahell.Report.fprintf Fmt.stderr "lhsort: %s; rhsort: %s@."
        (Z3.Sort.to_string @@ SMT.sort_of lexp) (Z3.Sort.to_string @@ SMT.sort_of rexp);*)
      { sort = bool_sort; expr = eq_expr lexp rexp }
    | And (phi1, phi2) ->
      let { expr = e1; _ } = to_z3 params out_expr phi1
      and { expr = e2; _ } = to_z3 params out_expr phi2 in
      { sort = bool_sort; expr = bool_and e1 e2 }

  let collect_vars : branch -> SMT.params =
    let rec aux acc = function
      | Atom (ops, Var (h, width)) ->
        let sort = SMT.bv_sort width in
        let expr = SMT.(const_expr sort (arg_sym h)) in
        (h, SMT.{ sort; expr }) :: acc
      | And (phi1, phi2) -> aux (aux acc phi1) phi2
      | _ -> acc
    in aux []

  let encode out_sort phi : SMT.problem =
    let phi = simplify_branch phi in
    SMT.encode out_sort (collect_vars phi) (fun args expr -> to_z3 args expr phi)
end
module ReprSkel = struct
  include Pattern(WordSkel)

  let compose_subskel subty_sym subterm_pos (subskel : t) (com_pos : Hole.t) =
    match subskel, com_pos with
    | Word { len; max_val; bits }, Hole.MHole ops ->
      let max_val = Hole.WordOps.max_input_val Z.(~$2 ** len - one) ops (Word max_val) in
      let min_len = if Z.numbits max_val >= len then Z.numbits max_val else len in
      let container_len = (match ops with
        | Hole.WordOps.Extract { low; high; ops } ->
          min_len + low
        | ops -> min_len) in
      let bits = [ WordSkel.(Atom (ops, Var (subterm_pos, len))) ] in
      Word { len = container_len; max_val; bits }
    | _, Hole.MHole ops when ops <> Hole.WordOps.Id ->
      (* cannot perform word ops on non-word stuff *)
      assert false
    | _ -> subst_in_hole com_pos (App { ty_sym = subty_sym; term_pos = subterm_pos })

  let inject_imm_opt ct_len ct_maxval (content : immediate option) (com_pos : Hole.t) =
    let rec split_mem_ctx = function
      | Hole.MHole ops -> Hole.MHole Id, ops
      | Hole.MHPtr { len; align; hole } ->
        let hole, ops = split_mem_ctx hole in
        Hole.MHPtr { len; align; hole }, ops
      | Hole.MHBlock { pos; hole } ->
        let hole, ops = split_mem_ctx hole in
        Hole.MHBlock { pos; hole }, ops
    in
    let mh, ops = split_mem_ctx com_pos in
    let bits = match content with
    | Some imm -> [ WordSkel.Atom (ops, Imm imm) ]
    | None -> [ WordSkel.Empty ] in
    subst_in_hole mh (Word { len = ct_len; max_val = ct_maxval; bits })

  let encode : t -> UnsolvedRepr.t =
    let encode_bits len bits : UnsolvedWord.t list =
      let out_sort = SMT.bv_sort len in
      List.map (fun phi ->
          WordSkel.collect_vars phi,
          WordSkel.encode out_sort phi)
        bits
    in let rec aux : t -> UnsolvedRepr.t = function
      | Empty -> Empty
      | App { ty_sym; term_pos } -> App { ty_sym; term_pos }
      | Word { len; max_val; bits } ->
        (match encode_bits len bits with
        | [] -> assert false
        | [ bits ] -> Word { len; max_val; bits }
        | l -> Union (List.map (fun bits -> UnsolvedRepr.Word { len; max_val; bits }) l))
      | Ptr { len; align; mpat; bits } ->
        (match encode_bits len bits with
        | [] -> assert false
        | [ bits ] -> Ptr { len; align; mpat = aux mpat; bits }
        | l ->
          let l = List.map
            (fun bits -> UnsolvedRepr.Ptr { len; align; mpat = aux mpat; bits })
            l in
          Union l)
      | Block l ->
        Block (List.map
          (function
            | None -> None
            | Some mp -> Some (aux mp))
          l)
      | Union l -> Union (List.map aux l)
    in aux

  (*let solve ~concrete : t -> SolvedRepr.t =
    fun x -> UnsolvedRepr.solve ~concrete @@ encode x*)
  (*let solve ~concrete : t -> SolvedRepr.t =
    let solve_bits ~concrete len bits : SolvedWord.t list =
      let out_sort = SMT.bv_sort len in
      List.map (fun phi ->
          (*Peahell.Report.fprintf Fmt.stderr "Solving %a on %d bits…@." WordSMT.pp_branch phi len;*)
          WordSkel.collect_vars phi,
          SMT.solve_problem ~concrete @@ WordSkel.encode out_sort phi)
        bits
      (*Peahell.Report.fprintf Fmt.stderr "Solved %a on %d bits, result: %a@."
        WordSMT.pp bits len
        Fmt.(list ~sep:(any " ∨ ") (pair ~sep:(any "") nop SMT.pp_interp)) res;*)
    in
    let rec aux : t -> SolvedRepr.t = function
      | Empty -> Empty
      | App { ty_sym; term_pos } -> App { ty_sym; term_pos }
      | Word { len; max_val; bits } ->
        (match solve_bits ~concrete len bits with
        | [] -> assert false
        | [ bits ] -> Word { len; max_val; bits }
        | l -> Union (List.map (fun bits -> SolvedRepr.Word { len; max_val; bits }) l))
      | Ptr { len; align; mpat; bits } ->
        (match solve_bits ~concrete len bits with
        | [] -> assert false
        | [ bits ] -> Ptr { len; align; mpat = aux mpat; bits }
        | l ->
          let l = List.map
            (fun bits -> SolvedRepr.Ptr { len; align; mpat = aux mpat; bits })
            l in
          Union l)
      | Block l ->
        Block (List.map (function
            | None -> None
            | Some mp -> Some (aux mp))
          l)
      | Union l ->
        Union (List.map aux l)
    in aux*)
end

(** Dot output *)
module Dot = struct
  let rec to_mempat : value -> SolvedRepr.t = function
    | VImm { width; bits } ->
      let bits = [], SMT.{ cases = []; default = smt_expr_of_imm bits } in
      Word { len = width; max_val = Z.((one + one) ** width - one); bits }
    | VPtr { width; align; pointee; bits } ->
      let bits = [], SMT.{ cases = []; default = smt_expr_of_imm bits } in
      Ptr { len = width; align; mpat = to_mempat pointee; bits }
    | VBlock l ->
      Block (List.map (fun v -> Some (to_mempat v)) l)

  let pp ppf v = SolvedRepr.Dot.pp ppf (to_mempat v)
end
