(* The parser *)

%{

open Input.Parse
open Syntax
   
let make_loc (startpos, endpos) = Location.Location (startpos, endpos)

type error +=
   | Unclosed of Location.loc * string * Location.loc * string
   | Doesnt_fit of Location.loc * Z.t
              
let unclosed opening_name opening_loc closing_name closing_loc =
  raise(Error(Unclosed(make_loc opening_loc, opening_name,
                       make_loc closing_loc, closing_name)))

(* Error reporting *)
let prepare_error = function
  | Error (Unclosed (opening_loc, _opening, closing_loc, closing)) ->
     Some (Report.errorf
        ~loc:closing_loc
        ~sub:[opening_loc]
        "Syntax error: '%s' expected" closing)
  | Error (Doesnt_fit (loc, i)) ->
     Some (Report.errorf
        ~loc
        "Syntax error: '%a' doesn't fit in an integer" Z.pp_print i)
  | _ -> None
let () = Report.register_report_of_exn prepare_error

  
%}

/* Tokens */
%token COMMA
%token EOF
%token DOT
%token COLON
%token EQUAL
%token <string> UIDENT
%token <string> LIDENT
%token LPAREN RPAREN
%token LANGLE RANGLE
%token LACCO RACCO
%token BAR
%token AMPERSAND
%token LET TYPE BITS SHOW FULL MATCHING
%token MINUSMINUS BACKSLASH ARROW
%token <Z.t> INT
%token <string> EXPECT
%token UNDERSCORE

%right AMPERSAND
%left BACKSLASH
%left BAR

/* Entry points */

%type <Syntax.command> command
%start program
%type <Syntax.program> program
%start toplevel_phrase
%type <Syntax.program> toplevel_phrase
%start expect_file
%type <(Syntax.program * int * int) list> expect_file

%start value_top
%type <Value.t> value_top

%%

program: list(item) EOF { $1}
toplevel_phrase: list(item) DOT {$1}
item: 
  | l = command EXPECT? { l }
;

expect_file: expect_item* EOF {$1}
%inline expect_item:
  | l = command+ EXPECT { l, $endofs(l), $endofs($2) }
;

command: show | value_declaration | type_declaration | matching { $1 };

show:
  | SHOW ty=type_expr { Show ty } 

value_declaration:
  LET 
  name = name
  COLON
  ty=type_expr
  EQUAL
  rhs=value
    { Let (name, ty, rhs) }
;

type_declaration:
  TYPE
  name = name
  parameters = parameters
  EQUAL
  definition = definition
    { Declaration {name; parameters; definition} }
;

definition:
  | BAR? l = separated_list(BAR,data_constructor) { Sum l }
  | BITS LANGLE width=width RANGLE values=valueset { Word {width; values} }
  | BITS LANGLE width=width error { unclosed "<" $loc($2) ">" $loc($4) }
;

width:
  | i=INT
    { try Z.to_int i
      with Z.Overflow -> raise (Error(Doesnt_fit(make_loc $loc(i),i)))
    }
  
valueset:
  | /*empty*/ {Full}
  | v=valueset_expr { v }

valueset_expr:
  | FULL { Full }
  | LPAREN e=valueset_expr RPAREN { e }
  | i=INT { Value i }
  | i1=INT MINUSMINUS i2=INT { Range (i1,i2) }
  | e1=valueset_expr AMPERSAND e2=valueset_expr { Union (e1,e2) }
  | e1=valueset_expr BACKSLASH e2=valueset_expr { Diff (e1,e2) }
;

data_constructor:
  | constructor=constr { {constructor; arguments = []} }
  | constructor=constr argument=simple_type_expr
    { {constructor; arguments=match argument with Tuple l -> l | _ -> [argument]} }
;

(* Type expressions *)
simple_type_expr:
  | name = name { Type_constructor { constructor = name; arguments = [] } }
  | AMPERSAND ty=simple_type_expr { Reference ty }
  | l=tuple(type_expr) { match l with | [x] -> x | l -> Tuple l }
;

type_expr:
  | ty=simple_type_expr {ty}
  | name = name arguments=tuple(type_expr) { Type_constructor { arguments; constructor=name } }
;

(* Matching *)
matching:
  | MATCHING pty=type_expr ARROW ety=type_expr
    LACCO clauses=clauses RACCO
    { Matching { pattern_type = pty; expr_type = ety; clauses } }

clauses: BAR? l = separated_list(BAR,clause) { l }
clause:
  | pattern=pattern_top ARROW expr=expr { {pattern;expr} }

pattern_top:
  | l=separated_nonempty_list(COMMA,pattern)
    { match l with [p] -> p | l -> PTuple l }
pattern:
  | UNDERSCORE { PAny }
  | p1=pattern BAR p2=pattern { POr (p1, p2) }
  | constructor=constr arguments=pattern_args
    { PConstr {constructor; arguments } }
  | l=tuple(pattern) { PTuple l }
  | var=name { PVar var }
pattern_args:
  | { [] }
  | LPAREN l=separated_list(COMMA,pattern) RPAREN
     { l }

expr:
  | constructor=constr arguments=expr_args
    { EConstr {constructor; arguments } }
  | l=tuple(expr) { ETuple l }
  | var=name { EVar var }
  | i=INT { EInt i }
expr_args:
  | { [] }
  | LPAREN l=separated_list(COMMA,expr) RPAREN
     { l }

value_top: v=value EOF { v }
value:
  | constructor=constr arguments=value_args
    { Value.VConstr {constructor; arguments } }
  | l=tuple(value) { Value.VTuple l }
  | i=INT { Value.VInt i }
  | AMPERSAND v=value { Value.VRef v }
value_args:
  | { [] }
  | v=value { match v with VTuple l -> l | v -> [v] }

parameters: 
  | /*empty*/ {[]}
  | l=tuple(name) {l};

tuple(r):
  | LPAREN l=separated_nonempty_list(COMMA,r) RPAREN {l};

name: LIDENT {$1};
constr: UIDENT {$1};
