open Types
open Target
open Syntax
open Value

let eprintf s = Peahell.Report.fprintf Fmt.stderr s

type 'a repr_cases =
  | TopCase
  | WordCases of (Z.t, Z.t -> 'a) Tag.KlassMap.t
  | SumCases of 'a Name.Map.t
type 'a reprs = { generic: 'a; specific: 'a repr_cases }
module type S = sig
  val pre_repr : Types.evaled_env -> Name.t -> Mem.ReprSkel.t reprs
  val repr : Types.evaled_env -> Name.t -> Mem.SolvedRepr.t reprs

  val knit : Types.evaled_env -> Name.t -> Hole.t -> Syntax.name -> Target.rexpr

  (** [frog tau h l x] émet le code qui discrimine la repr située en focus(h, x)
    suivant les cas de l; on garantit que l est exhaustive + typage se passe bien
    tau = type de x *)
  val frog : Types.evaled_env -> Name.t -> Hole.t ->
    Target.dtree Tag.Map.t -> Syntax.name -> Target.dtree
end

module type Eval = sig
  val to_pre_repr : Types.evaled_env -> Name.t -> Value.t -> SMT.problem
  val to_repr : Types.evaled_env -> Name.t -> Value.t -> Mem.t
  val match_repr : Types.evaled_env -> Name.t -> Syntax.pattern list -> Mem.t -> int
end

module type Full = sig
  include S
  include Eval
end

let tag_to_word ~variants k : Z.t =
  match List.concat (List.mapi (fun i (k',_) -> if k' = k then [i] else []) variants)
  with 
  | [i] -> Z.of_int i
  | _ ->
    eprintf "Could not find constructor %s in type %a@." k Types.pp_evaled (Sum variants);
    assert false

let sort_cases l =
  let cmp (z,_) (z',_) = Z.compare z z' in
  List.sort cmp l 

let wrap_case_lhs (z, dt) = (Target.LWord z), dt

let rec subty_at (h : Hole.t) (ty_env : Types.evaled_env) ty_sym =
  match h, Types.find_and_unroll ty_sym ty_env with
  | Hole, _ -> ty_sym
  | HRef h, Ref ty -> subty_at h ty_env ty
  | HTup { arity; pos; hole }, Product l ->
    subty_at hole ty_env (List.nth l pos)
  | HConstr { constr; arity; pos; hole }, Sum l ->
    subty_at hole ty_env (List.nth (List.assoc constr l) pos)
  | _ -> assert false
let rec focus_ty (h : Hole.t) (ty_env : Types.evaled_env) ty_sym =
  match h, Types.find_and_unroll ty_sym ty_env with
  | Hole, ty -> ty
  | HRef h, Ref ty -> focus_ty h ty_env ty
  | HTup {arity; pos; hole}, Product l ->
    focus_ty hole ty_env (List.nth l pos)
  | HConstr {constr; arity; pos; hole}, Sum l ->
    focus_ty hole ty_env (List.nth (List.assoc constr l) pos)
  | _, ty ->
    eprintf "Mismatched holexpr types: %a ≠ %a@." Hole.pp h pp_evaled ty;
    assert false
let rec focus_val (h:Hole.t) v = match h, v with
  | Hole, _ -> v
  | HRef h, VRef v -> focus_val h v
  | HTup { arity; pos; hole }, VTuple l when List.length l = arity ->
    focus_val hole (List.nth l pos)
  | HConstr { constr; arity; pos; hole }, VConstr { constructor; arguments }
    when constr = constructor && List.length arguments = arity ->
    focus_val hole (List.nth arguments pos)
  | _ ->
    eprintf "Mismatched holexpr types.";
    assert false

module Uniform = struct
  (*
    1. tuple = &(bloc [tag=tuplen; field0; ... fieldn-1])
    2. constr = &(bloc [tag=constr; field0; ... fieldn-1])
  *)
  let field_width = 1

  let repr = fun _ -> assert false
  let pre_repr = fun _ -> assert false

  let rec focus_repr h x : Target.rexpr = match h with
    | Hole.Hole -> x
    | HRef h' -> focus_repr h' (RDeref x)
    | HTup {arity; pos; hole} ->
      	focus_repr hole (RAccess(RDeref x, field_width * (pos+1), field_width))
    | HConstr {constr; arity; pos; hole} ->
      	focus_repr hole (RAccess(RDeref x, field_width * (pos+1), field_width))

  let tag_repr h x : rexpr = RAccess(RDeref (focus_repr h x), 0, field_width)

  let frog ty_env ty_sym h l x =
    let ty = focus_ty h ty_env ty_sym in
    match unroll_vars ty_env ty, Tag.classify l with
    | _, `Any dt -> dt
    | Word _, `Ints (cases, default) ->
      let cases = List.map wrap_case_lhs (sort_cases cases) in
      Switch { expr = focus_repr h (RVar x); cases; default}
    | Sum variants, `Constrs (cases, default) ->
      let cases =
        List.map wrap_case_lhs @@
        sort_cases @@
        List.map (function (k, dt) -> (tag_to_word ~variants k), dt) cases
      in
      Switch { expr = tag_repr h (RVar x); cases; default }
    | _, _ -> assert false

	let knit ty_env ty_sym h x = focus_repr h (RVar x)
    (*eprintf "Rebuilding %a : %a@." Hole.pp h pp_rexpr res;*)

  (* version eval *)
  let rec to_repr ty_env ty_sym v : Mem.t =
    match find_and_unroll ty_sym ty_env, v with
    | Word _, Value.VInt z -> VImm { width = 64; bits = Word z }
    | Ref sym, VRef v' ->
      Mem.VPtr { width = 64; align = 0;
        pointee = to_repr ty_env sym v'; bits = Word Z.zero }
    | Product l, VTuple vv ->
      let rr = List.map2 (to_repr ty_env) l vv in
      Mem.VPtr { width = 64; align = 0; pointee = VBlock rr; bits = Word Z.zero }
    | Sum variants, VConstr { constructor; arguments } ->
      let arg_tys = List.assoc constructor variants in
      Mem.VPtr { width = 64; align = 0;
        pointee = VBlock (
          VImm { width = 64; bits = Word (tag_to_word ~variants constructor) }
          :: List.map2 (to_repr ty_env) arg_tys arguments);
        bits = Word Z.zero }
    | _ -> assert false

  let to_pre_repr = fun _ -> assert false

  let rec bin_pat_match : pattern -> Mem.t -> bool = fun p r -> match p, r with
    | PAny, _ -> true
    | PVar x, _ -> true
    | POr (p1, p2), _ -> bin_pat_match p1 r || bin_pat_match p2 r
    | PInt z, VImm { bits = Word z'; _ } -> z = z'
    | PRef p', VPtr { pointee; _ } -> bin_pat_match p' pointee
    | PTuple l, VPtr { pointee = VBlock reprs; _ } ->
        List.for_all2 bin_pat_match l reprs
    | PConstr { constructor; arguments }, VPtr { pointee = VBlock (_ :: rr); _ } ->
        List.for_all2 bin_pat_match arguments rr
    | _ -> assert false

  let match_repr ty_env ty_sym pats r : int = assert false (* TODO *)
end

module Camlike = struct
  (*
    1. tuple = &(bloc [tag=tuplen; field0; ... fieldn-1])
    2. constr = &(bloc [tag=constr; field0; ... fieldn-1])
    3. constr unit = Z.t(K) . 1
  *)
  let field_width = 1
  let repr = fun _ -> assert false
  let pre_repr = fun _ -> assert false

  let rec focus_repr h x : Target.rexpr = match h with
    | Hole.Hole -> x
    | HRef h' -> focus_repr h' (RDeref x)
    | HTup {arity; pos; hole} ->
      	focus_repr hole (RAccess(RDeref x, field_width * (pos+1), field_width))
    | HConstr {constr; arity; pos; hole} ->
      	focus_repr hole (RAccess(RDeref x, field_width * (pos+1), field_width))

  let tag_rexpr h x : rexpr = RAccess(RDeref (focus_repr h x), 0, field_width)

  let split_variants (variants: (Name.t * 'a list) list) =
    let units, others =
      List.partition (function (_, args) -> args = []) variants in
    units, others

  let make_cases ~variants branches =
    let in_variants (n, _) = List.mem_assoc n variants in
    let cases = List.filter in_variants branches in
    cases, (List.length cases = List.length variants)
    

  let tag_to_word ~variants k =
    let i = tag_to_word ~variants k in
    Z.((i lsl 1) lor ~$1)
  
  let frog ty_env ty_top h l x =
    (* eprintf "Destructing %s %a@." x pp_holexpr h; *)
    let ty = focus_ty h ty_env ty_top in
    match unroll_vars ty_env ty, Tag.classify l with 
    | _, `Any dt -> dt
    | Word _, `Ints (cases, default) ->
      let cases = List.map wrap_case_lhs (sort_cases cases) in
      Switch {expr = focus_repr h (RVar x); cases; default}
    | Sum variants, `Constrs (branches, default) ->
      let unit_variants, other_variants = split_variants variants in
      let cases_of_variants ~variants =
        let cases, full = make_cases ~variants branches in
        let cases =
          sort_cases @@
          List.map (function (k, dt) -> (tag_to_word ~variants k), dt) @@
          cases
        in
        cases, full        
      in
      let unit_cases, unit_full = cases_of_variants ~variants:unit_variants in
      let other_cases, other_full = cases_of_variants ~variants:other_variants in
      let unit_rexpr = focus_repr h (RVar x) in
      let other_rexpr = tag_rexpr h (RVar x) in
      begin match unit_cases, unit_full, other_cases, other_full with
        | cases, _, [], _ ->
          Target.switch ~default ~expr:unit_rexpr (List.map wrap_case_lhs cases)
        | [], true, cases, _ ->
          (* Beware: if there are unit variants in this type (but none
             were in the branches), compiling the switch this way is unsound,
             as it involves dereferencing something that might not be a pointer!
             Therefore, we verify above if `unit_full` is true
          *)
          Target.switch ~default ~expr:other_rexpr (List.map wrap_case_lhs cases)
        | _ ->
          let is_immediate = RAnd(focus_repr h (RVar x), Z.of_int 0b1) in
          let unit_switch =
            let default = if unit_full then None else default in
            let expr = unit_rexpr in
            Target.switch ~default ~expr (List.map wrap_case_lhs unit_cases)
          in
          let other_switch = 
            let default = if other_full then None else default in
            let expr = other_rexpr in
            Target.switch ~default ~expr (List.map wrap_case_lhs other_cases)
          in          
          Target.switch
            ~default:None
            ~expr:is_immediate
            [
              (Target.LWord (Z.of_int 1), unit_switch);
              (Target.LWord (Z.of_int 0), other_switch);
            ]            
      end 
    | _ -> assert false

	let knit ty_env ty_sym h x = focus_repr h (RVar x)
end

module Composable = struct

  type 'a discriminant =
    { position : Mem.Hole.t; min_width : int option;
      split : ('a, Mem.immediate option) Tag.KlassMap.t } (* None = tag placeholder *)
  module type Spec = sig
    val recompose : Types.evaled_env -> Name.t -> Hole.Head.t -> Mem.Hole.t
    val tagsets_of_word : Types.word_spec -> Z.t Tag.Klass.t list
    val discrs_of_word : Types.word_spec -> Z.t discriminant list
    val discrs_of_sum : Types.evaled_env -> Types.sum_spec -> Name.t discriminant list
  end
  module Make (R : Spec) = struct
    (* Distinguishability stuff *)
    let disting_skels_of_tagset wspec tagset (tag : Z.t option) =
      let max_tag_val = Tag.Klass.max_elt tagset in
      let max_tag_width = Z.numbits max_tag_val in
      List.filter_map
        (fun { position; min_width; split } ->
          match Tag.KlassMap.find_opt tagset split with
          | Some None ->
            let width = (match min_width with
              | Some width -> width
              | None -> max_tag_width)
            and tag_imm = (match tag with
              | Some z -> Some (Mem.Word z)
              | None -> None) in
            Some (Mem.ReprSkel.inject_imm_opt width max_tag_val tag_imm position)
          | Some (Some imm) ->
            let max_val = Mem.as_z imm in
            let width = (match min_width with
              | Some width -> width
              | None -> Z.numbits max_val) in
            Some (Mem.ReprSkel.inject_imm_opt width max_val (Some imm) position)
          | None -> None)
      (R.discrs_of_word wspec)

    let disting_skels_of_constr ty_env sumspec constr =
      let single_constr = Tag.Klass.singleton_constrs constr in
      List.filter_map
        (fun { position; min_width; split } ->
          match Tag.KlassMap.find_opt single_constr split with
          | Some None -> assert false (* only int tags may be used as discrs *)
          | Some (Some imm) ->
            let width = (match min_width with
              | Some width -> width
              | None ->
                let max_val = Mem.Hole.max_input_val (Mem.as_z imm) position imm in
                if max_val = Z.zero then 1 else Z.numbits max_val) in
            Some (Mem.ReprSkel.inject_imm_opt width (Mem.as_z imm) (Some imm) position)
          | None -> None)
        (R.discrs_of_sum ty_env sumspec)

    let repr_skels (ty_env : Types.evaled_env) root_sym :
    (Mem.ReprSkel.t reprs) Name.Map.t =
      let rec aux (acc : (Mem.ReprSkel.t reprs) Name.Map.t) term_pos ty_sym =
        if Name.Map.mem ty_sym acc then acc else
        let acc = (* add recursive repr sym *)
          Name.Map.add ty_sym
            { generic = Mem.ReprSkel.App { ty_sym; term_pos };
              specific = TopCase }
            acc in

        let to_generic_skels = Name.Map.map (fun skel -> skel.generic) in
        let inter_skels acc = Mem.ReprSkel.inter_list (to_generic_skels acc)
        and union_skels acc = Mem.ReprSkel.union_list (to_generic_skels acc) in

        let compose_skel acc subty h = Mem.ReprSkel.compose_subskel
          subty Hole.(subst term_pos @@ Head.as_context h)
          (Name.Map.find subty acc).generic (R.recompose ty_env ty_sym h)

        and extend_term_pos h = Hole.(subst term_pos Head.(as_context h)) in

        match Types.find_and_unroll ty_sym ty_env with
        | Types.Word wspec ->
          let do_one_tagset (gen_cases, spec_cases) tagset =
            let new_spec_case =
              fun z -> inter_skels acc @@ disting_skels_of_tagset wspec tagset (Some z)
            and new_gen_case =
              inter_skels acc @@ disting_skels_of_tagset wspec tagset None in
            new_gen_case :: gen_cases,
            Tag.KlassMap.add tagset new_spec_case spec_cases
          in
          let gen_cases, spec_cases = List.fold_left do_one_tagset
            ([], Tag.KlassMap.empty_intm)
            (R.tagsets_of_word wspec) in
          Name.Map.add ty_sym
            { generic = union_skels acc gen_cases;
              specific = WordCases spec_cases }
            acc
        | Types.Ref subty ->
          let ctx = Hole.Head.Ref in
          let acc = aux acc (extend_term_pos ctx) subty in
          Name.Map.add ty_sym
            { generic = compose_skel acc subty ctx;
              specific = TopCase }
            acc
        | Types.Product subtys ->
          let arity = List.length subtys in
          let ctxs = List.mapi (fun pos subty -> Hole.Head.Tuple { pos; arity }) subtys in
          let acc = List.fold_left2 aux acc (List.map extend_term_pos ctxs) subtys in
          Name.Map.add ty_sym
            { generic = inter_skels acc @@ List.map2
                (fun subty ctx -> compose_skel acc subty ctx)
                subtys ctxs;
              specific = TopCase }
            acc
        | Types.Sum sumspec ->
          let do_one_constr (acc, cases) (constr, subtys) =
            let arity = List.length subtys in
            let ctxs = List.mapi
              (fun pos subty -> Hole.Head.Constr { constr; pos; arity })
              subtys in
            let acc = List.fold_left2 aux acc (List.map extend_term_pos ctxs) subtys in
            let new_case = inter_skels acc @@
              disting_skels_of_constr ty_env sumspec constr @
              List.map2 (fun subty ctx -> compose_skel acc subty ctx) subtys ctxs in
            (*Peahell.Report.fprintf Fmt.stderr "Repr of constr %s is %a@."
              constr Mem.ReprSkel.pp new_case;*)
            acc, Name.Map.add constr new_case cases
          in
          let acc, cases = List.fold_left do_one_constr (acc, Name.Map.empty) sumspec in
          let res = Name.Map.add ty_sym
            { generic = union_skels acc (snd @@ List.split @@ Name.Map.bindings cases);
              specific = SumCases cases }
            acc in
          (*Peahell.Report.fprintf Fmt.stderr "Generic repr is %a@."
            Mem.SMTSkel.pp (Name.Map.find ty_sym res).generic;*)
          res
        | _ -> assert false
      in
      aux Name.Map.empty Hole.Hole root_sym


    let do_repr_skel f { generic; specific } =
      let generic = f generic
      and specific = match specific with
        | TopCase -> TopCase
        | WordCases kmap ->
          let kmap = Tag.KlassMap.map
            (fun z_to_skel ->
              fun z -> f (z_to_skel z))
            kmap in
          WordCases kmap
        | SumCases m ->
          SumCases (Name.Map.map f m)
      in
      { generic; specific }
    (*let solve_repr_skel ~concrete { generic; specific } =
      let generic = Mem.ReprSkel.solve ~concrete generic
      and specific = match specific with
        | TopCase -> TopCase
        | WordCases kmap ->
          let kmap = Tag.KlassMap.map
            (fun z_to_skel ->
              fun z -> Mem.ReprSkel.solve ~concrete (z_to_skel z))
            kmap in
          WordCases kmap
        | SumCases m ->
          SumCases (Name.Map.map (Mem.ReprSkel.solve ~concrete) m)
      in
      { generic; specific }*)

    let pre_repr ty_env ty_sym =
      Name.Map.find ty_sym (repr_skels ty_env ty_sym)
    let repr ty_env ty_sym =
      do_repr_skel
        Mem.(fun s -> UnsolvedRepr.solve ~concrete:false @@ ReprSkel.encode s)
        (Name.Map.find ty_sym (repr_skels ty_env ty_sym))

    let to_pre_repr ty_env ty_sym v = assert false

    let to_repr ty_env ty_sym v =
      let tag_of ty v = match ty, v with
        | Types.Word _, VInt z -> Tag.Int z
        | Types.Sum _, VConstr { constructor; _ } -> Tag.Constr constructor
        | _ -> Tag.Top
      in
      let reprs_env = Name.Map.map
        (do_repr_skel
          Mem.(fun s -> UnsolvedRepr.solve ~concrete:true @@ ReprSkel.encode s))
        (repr_skels ty_env ty_sym) in
      let repr_env = Name.Map.map
        (fun { generic; specific } ->
          fun tag -> match tag, specific with
            | Tag.Top, _ -> generic
            | Tag.Int z, WordCases kmap -> Tag.KlassMap.findtag z kmap z
            | Tag.Constr k, SumCases m -> Name.Map.find k m
            | _ -> assert false)
        reprs_env in
      let ty_repr = match Name.Map.find_opt ty_sym repr_env with
        | Some r -> r
        | None -> failwith ty_sym in
      let tag = tag_of (find_and_unroll ty_sym ty_env) v in
      Mem.SolvedRepr.to_memval
        (fun h -> subty_at h ty_env ty_sym)
        (fun term_pos ->
          let subty = focus_ty term_pos ty_env ty_sym
          and subval = focus_val term_pos v in
          tag_of subty subval)
        repr_env
        (ty_repr tag)

        
    let match_repr ty pats rv = assert false (* TODO *)
   
    let full_recompose = Hole.by_induction Mem.Hole.subst R.recompose
    
    let knit ty_env ty_sym h x = Mem.Hole.focus (full_recompose ty_env ty_sym h) (RVar x)

    let compare_lhs lhs1 lhs2 = match lhs1, lhs2 with
      | LWord z1, LWord z2 -> Z.compare z1 z2
      | LName n1, LName n2 -> String.compare n1 n2
      | x, y -> compare x y
    module LhsMap = Map.Make(struct type t = Target.lhs let compare = compare_lhs end)
    let imm_to_lhs = function
      | Mem.True -> Target.LWord (Z.one)
      | Mem.False -> Target.LWord (Z.zero)
      | Mem.Word z -> Target.LWord z
      | Mem.Symbol cst -> Target.LName cst
    and cmp (lhs1, _) (lhs2, _) = compare_lhs lhs1 lhs2
    let frog ty_env ty_sym h tagmap x =
      let prefocus = full_recompose ty_env ty_sym h in
      let ty = focus_ty h ty_env ty_sym in
      let rec aux : type a. a Tag.kind -> a discriminant list -> (a -> a Tag.Klass.t) ->
        (a * dtree) list -> dtree option -> a Tag.Klass.t -> dtree =
        fun kind discrs set_of_tag cases default bounding_tagset ->
        match cases, default with
        | [], None -> Fail
        | [ _, dtree ], None | [], Some dtree -> dtree
        | [ tag, dtree ], Some _ ->
          let other_tags = Tag.Klass.remove tag bounding_tagset in
          let relevant_discrs = List.sort
            (fun (pos, _) (pos', _) -> Mem.Hole.cmp pos' pos) @@
            List.filter_map
              (fun { position; split; _ } ->
                let relevant_tagsets = List.filter (fun (tagset, rhs) ->
                    Tag.Klass.(not @@ is_empty @@ inter other_tags tagset))
                  (Tag.KlassMap.bindings split) in
                if relevant_tagsets = [] then None
                else match Tag.KlassMap.find_opt (set_of_tag tag) split with
                | Some None -> (match kind with
                  | Tag.TInt -> Some (position, Target.LWord tag)
                  | Tag.TConstr -> assert false)
                | Some (Some imm) -> Some (position, imm_to_lhs imm)
                | None -> None)
              discrs in
          List.fold_left
            (fun dtree (pos, lhs) ->
              let expr = Mem.Hole.focus (Mem.Hole.subst prefocus pos) (RVar x) in
              switch ~default ~expr [ lhs, dtree ])
            dtree relevant_discrs
        | cases, _ ->
          (* must distinguish between at least two cases,
             through at least one discr with two distinct lhs *)
          let relevant_discrs = List.filter_map
            (fun { position; split; _ } ->
              let split = Tag.KlassMap.bindings split in
              let filtered_by (tagset, rhs) = List.filter_map
                ((fun (tag, _) -> if Tag.Klass.mem tag tagset
                  then Some (match rhs, kind with
                    | Some imm, _ -> imm | None, Tag.TInt -> Mem.Word tag | _ -> assert false)
                  else None) : a * dtree -> Mem.immediate option)
                cases in
              let distinct_lhs = List.sort_uniq compare @@
                List.concat_map filtered_by split in
              if List.length distinct_lhs > 1 then Some (position, split) else None)
            discrs in
          (* TODO choose a min discr *)
          let position, split =
            let ordered_discrs = List.sort
              (fun (pos, _) (pos', _) -> Mem.Hole.cmp pos' pos)
              relevant_discrs in
            List.nth ordered_discrs 0 in

          let undist_tagset = List.fold_left
            (fun acc (tagset, _) -> Tag.Klass.diff acc tagset)
            bounding_tagset split in
          let undist_cases = List.filter
            (fun (tag, _) -> Tag.Klass.mem tag undist_tagset)
            cases in
          let undist_default =
            if Tag.Klass.is_empty undist_tagset then None
            else Some (aux kind discrs set_of_tag undist_cases default undist_tagset) in

          let group_by_lhs : (a Tag.Klass.t * Mem.immediate option) list ->
            (* lhs -> filtered classes, bounding tagset of filtered tags *)
            (a Tag.Klass.t list * a Tag.Klass.t) LhsMap.t =
            let add_to_lhsmap lhs (klass, tagset) m = match LhsMap.find_opt lhs m with
              | Some (klasses, tagset') ->
                let klasses =
                  if List.mem klass klasses then klasses else klass :: klasses in
                LhsMap.add lhs (klasses, Tag.Klass.union tagset tagset') m
              | None ->
                LhsMap.add lhs ([ klass ], tagset) m in
            List.fold_left
              (fun acc ((klass : a Tag.Klass.t), discr_val) ->
                let tags = List.filter_map
                  (fun (tag, _) -> if Tag.Klass.mem tag klass then Some tag else None)
                  cases in
                let tagset : a Tag.Klass.t = match kind with
                  | Tag.TInt -> Tag.Klass.ints_of_list tags
                  | Tag.TConstr -> Tag.Klass.constrs_of_list tags in
                match discr_val with
                | Some imm ->
                  add_to_lhsmap (imm_to_lhs imm) (klass, tagset) acc
                | None -> (* may result in several different lhs *)
                  match kind with
                  | Tag.TInt ->
                    List.fold_left
                      (fun acc tag -> add_to_lhsmap (Target.LWord tag) (klass, tagset) acc)
                      acc tags
                  | Tag.TConstr -> assert false)
              LhsMap.empty in

          let branches = List.sort cmp @@ List.filter_map
            (fun (lhs, (klasses, full_tagset)) ->
              let cases = List.filter
                (fun (tag, _) -> Tag.Klass.mem tag full_tagset)
                cases in
              if cases = [] then None
              else Some (lhs, aux kind discrs set_of_tag cases undist_default full_tagset))
            (LhsMap.bindings @@ group_by_lhs split)

          and expr = Mem.Hole.focus (Mem.Hole.subst prefocus position) (RVar x) in
          switch ~default ~expr branches
        in
        match Types.unroll_vars ty_env ty, Tag.classify tagmap with
          | _, `Any deft -> deft
          | Word wspec, `Ints (cases, default) ->
            let set_of_tag =
              fun z -> List.find (Tag.Klass.mem z) (R.tagsets_of_word wspec) in
            aux Tag.TInt (R.discrs_of_word wspec) set_of_tag cases default (Tag.Klass.starter_ints wspec)
          | Sum l, `Constrs (cases, default) ->
            let set_of_tag = fun constr -> Tag.Klass.singleton_constrs constr in
            aux Tag.TConstr (R.discrs_of_sum ty_env l) set_of_tag cases default (Tag.Klass.starter_constrs l)
          | _ -> assert false
  end
end
