type name = string

type 'a constructor = {
  constructor : name ;
  arguments : 'a list ;
}

and type_expr =
  | Type_constructor of type_expr constructor
  | Tuple of type_expr list
  | Reference of type_expr

and definition =
  | Sum of type_expr constructor list
  | Word of { width: int; values: valueset }

and valueset =
  | Full
  | Value of Z.t
  | Range of Z.t * Z.t
  | Union of valueset * valueset
  | Diff of valueset * valueset
 
and declaration = {
  name : name;
  parameters : name list; 
  definition : definition;
}

and matching = {
  pattern_type : type_expr ;
  expr_type : type_expr ;
  clauses : clause list ;
}

and clause = {
  pattern : pattern ;
  expr : expr ;
}

and pattern =
  | PVar of name
  | PConstr of pattern constructor
  | PTuple of pattern list
  | PAny
  | PRef of pattern
  | POr of pattern * pattern
  | PInt of Z.t

and expr =
  | EVar of name
  | EConstr of expr constructor
  | ETuple of expr list
  | EInt of Z.t

and value =
  | VInt of Z.t
  | VRef of value
  | VTuple of value list
  | VConstr of value constructor

and command =
  | Declaration of declaration
  | Show of type_expr
  | Let of name * type_expr * value
  | Matching of matching

and program = command list

let rec pp_expr fmt e = match e with
  | EVar x -> Fmt.pf fmt "%s" x
  | EConstr { constructor; arguments } ->
    Fmt.pf fmt "%s(%a)" constructor
      (Fmt.list ~sep:Fmt.comma pp_expr) arguments
  | ETuple l -> Fmt.pf fmt "< %a >" (Fmt.list ~sep:Fmt.comma pp_expr) l
  | EInt z -> Fmt.pf fmt "%a" Z.pp_print z

let rec pp_pattern fmt p = match p with
  | PVar x -> Fmt.pf fmt "%s" x
  | PConstr { constructor; arguments = [] } ->
    Fmt.pf fmt "%s" constructor
  | PConstr { constructor; arguments } ->
    Fmt.pf fmt "%s(%a)" constructor (Fmt.list ~sep:Fmt.comma pp_pattern) arguments
  | PTuple l -> Fmt.pf fmt "< %a >" (fun fmt -> List.iter (pp_pattern fmt)) l
  | PAny -> Fmt.pf fmt "_"
  | PRef p' -> Fmt.pf fmt "&(%a)" pp_pattern p'
  | POr(p1, p2) -> Fmt.pf fmt "%a | %a" pp_pattern p1 pp_pattern p2
  | PInt z -> Fmt.pf fmt "%a" Z.pp_print z
