type t = Top | Int of Z.t | Constr of Syntax.name
module M = struct
  type nonrec t = t
  let compare = Stdlib.compare
end
module Set = CCSet.Make(M)
module Map = CCMap.Make(M)


type _ kind =
  | TInt : Z.t kind
  | TConstr : Name.t kind

module Klass = struct
  type tagt = t
  module IntSet = Diet.Make(Z)
  module ConstrSet = CCSet.Make(String)
  type _ t =
    | Ints : IntSet.t -> Z.t t
    | Constrs : ConstrSet.t -> Syntax.name t

  let pp : type a. _ -> a t -> unit = fun fmt s ->
    match s with
    | Ints s -> IntSet.pp fmt s
    | Constrs s -> ConstrSet.pp Name.pp fmt s

  let max_elt : type a. a t -> a = function
    | Ints s -> IntSet.(Interval.y @@ max_elt s)
    | Constrs _ -> assert false

  let empty_ints = Ints (IntSet.empty)
  let empty_constrs = Constrs (ConstrSet.empty)

  let starter_ints ({ width; _ } : Types.word_spec) =
    let range = IntSet.Interval.make Z.zero Z.((one + one) ** width - one) in
    Ints IntSet.(add range empty)
  let starter_constrs (l : Types.sum_spec) =
    Constrs (ConstrSet.of_list (fst (List.split l)))

  let is_empty : type a. a t -> bool = function
    | Ints s -> IntSet.is_empty s
    | Constrs s -> ConstrSet.is_empty s

  let mem : type a. a -> a t -> bool = fun x s -> match s with
    | Ints s -> IntSet.mem x s
    | Constrs s -> ConstrSet.mem x s
  let memtag : type a. tagt -> a t -> bool = fun t s -> match s, t with
    | Ints s, Int c -> IntSet.mem c s
    | Constrs s, Constr k -> ConstrSet.mem k s
    | _ -> assert false

  let add : type a. a -> a t -> a t = fun x s -> match s with
    | Ints s -> Ints IntSet.(add (Interval.make x x) s)
    | Constrs s -> Constrs (ConstrSet.add x s)
  let addtag : type a. tagt -> a t -> a t = fun t s -> match s, t with
    | Ints s, Int c -> Ints IntSet.(add (Interval.make c c) s)
    | Constrs s, Constr k -> Constrs (ConstrSet.add k s)
    | _ -> assert false
  let remove : type a. a -> a t -> a t = fun x s -> match s with
    | Ints s -> Ints IntSet.(remove (Interval.make x x) s)
    | Constrs s -> Constrs (ConstrSet.remove x s)

  let singleton_ints x = add x empty_ints
  let singleton_constrs x = add x empty_constrs
  
  let ints_of_list = List.fold_left (fun acc x -> add x acc) empty_ints
  let constrs_of_list = List.fold_left (fun acc x -> add x acc) empty_constrs

  let equal : type a. a t -> a t -> bool = fun s1 s2 -> match s1, s2 with
   | Ints s1, Ints s2 -> IntSet.equal s1 s2
   | Constrs s1, Constrs s2 -> ConstrSet.equal s1 s2
   | _ -> assert false

  let union : type a. a t -> a t -> a t = fun s1 s2 -> match s1, s2 with
   | Ints s1, Ints s2 -> Ints (IntSet.union s1 s2)
   | Constrs s1, Constrs s2 -> Constrs (ConstrSet.union s1 s2)
   | _ -> assert false

  let inter : type a. a t -> a t -> a t = fun s1 s2 -> match s1, s2 with
   | Ints s1, Ints s2 -> Ints (IntSet.inter s1 s2)
   | Constrs s1, Constrs s2 -> Constrs (ConstrSet.inter s1 s2)
   | _ -> assert false

  let diff : type a. a t -> a t -> a t = fun s1 s2 -> match s1, s2 with
   | Ints s1, Ints s2 -> Ints (IntSet.diff s1 s2)
   | Constrs s1, Constrs s2 -> Constrs (ConstrSet.diff s1 s2)
   | _ -> assert false
end

module KlassMap = struct
  type tagt = t
  module IntSetMap = CCMap.Make(struct type t = Z.t Klass.t
                                       let compare = Stdlib.compare end)
  module ConstrSetMap = CCMap.Make(struct type t = Syntax.name Klass.t
                                          let compare = Stdlib.compare end)
  type (_, 'b) t =
    | IntMap : 'b IntSetMap.t -> (Z.t, 'b) t
    | ConstrMap : 'b ConstrSetMap.t -> (Name.t, 'b) t
  
  let mem : type a b. a Klass.t -> (a, b) t -> bool = fun s m -> match m with
    | IntMap m -> IntSetMap.mem s m
    | ConstrMap m -> ConstrSetMap.mem s m

  let find : type a b. a Klass.t -> (a, b) t -> b = fun s m -> match m with
    | IntMap m -> IntSetMap.find s m
    | ConstrMap m -> ConstrSetMap.find s m

  let find_opt : type a b. a Klass.t -> (a, b) t -> b option = fun s m -> match m with
    | IntMap m -> IntSetMap.find_opt s m
    | ConstrMap m -> ConstrSetMap.find_opt s m

  let bindings : type a b. (a, b) t -> (a Klass.t * b) list = function
    | IntMap m -> IntSetMap.bindings m
    | ConstrMap m -> ConstrSetMap.bindings m

  let findtag : type a b. a -> (a, b) t -> b = fun tag m ->
    let _, res = List.find (fun (klass, _) -> Klass.mem tag klass) (bindings m) in
    res

  let map : type a b c. (b -> c) -> (a, b) t -> (a, c) t = fun f m -> match m with
    | IntMap m -> IntMap (IntSetMap.map f m)
    | ConstrMap m -> ConstrMap (ConstrSetMap.map f m)

  let empty_intm = IntMap (IntSetMap.empty)
  let empty_constrm = ConstrMap (ConstrSetMap.empty)

  let add : type a b. a Klass.t -> b -> (a, b) t -> (a, b) t
  = fun k v m -> match m with
    | IntMap m -> IntMap (IntSetMap.add k v m)
    | ConstrMap m -> ConstrMap (ConstrSetMap.add k v m)
end

let pp fmt = function
  | Top -> Fmt.pf fmt "⊤"
  | Int z -> Z.pp_print fmt z
  | Constr k -> Fmt.pf fmt "%s" k

let classify m =
  let aux tag dt (constrs, ints, any) = match tag with
    | Constr k -> ((k, dt) :: constrs, ints, any)
    | Int i -> (constrs, (i, dt) :: ints, any)
    | Top -> (constrs, ints, Some dt)
  in
  match Map.fold aux m ([], [], None) with
  | [], [], None -> `None
  | [], [], Some dt -> `Any dt
  | l, [], any -> `Constrs (l, any)
  | [], l, any -> `Ints (l, any)
  | mixed -> `Mixed mixed
