(** Compilation target *)

type name = Syntax.name
type word = Z.t

type rcmp =
  | Eq | Neq | Leq | Lt | Geq | Gt
let cmp_to_z = function
  | Eq -> Z.equal | Neq -> (fun z z' -> not Z.(z = z'))
  | Leq -> Z.leq | Geq -> Z.geq | Lt -> Z.lt | Gt -> Z.gt
and cmp_to_str = function
  | Eq -> "=" | Neq -> "≠"
  | Leq -> "≤" | Lt -> "<"
  | Geq -> "≥" | Gt -> ">"

type rarith =
  | Add | Sub | Mul | Div | Mod
let arith_to_z = function
  | Add -> Z.add | Sub -> Z.sub | Mul -> Z.mul | Div -> Z.cdiv | Mod -> Z.rem
and arith_to_str = function
  | Add -> "+" | Sub -> "-" | Mul -> "×" | Div -> "÷" | Mod -> "%"

type rexpr =
  | RVar of name
  | RWord of word
  | RDeref of rexpr
  (* operations on "slices" *)
  | RAccess of rexpr * int * int (* second int is width *)
  | RNotEmpty of rexpr
  | RHSlice of { slice: rexpr; rel_start: int; rel_end: int } 
  (* "vertical slice": select the first k bytes from each n-byte offset *)
  | RVSlice of { slice: rexpr; for_each: int; select: int }
  (* arith/bitwise/logical ops *)
  | RLShift of rexpr * int | RRShift of rexpr * int
  | RAnd of rexpr * word | RXor of rexpr * word
  | RNot of rexpr | RLogAnd of rexpr * rexpr | RLogOr of rexpr * rexpr
  | RCmp of rexpr * rcmp * Z.t
  | RArith of rexpr * rarith * Z.t

let rec simplify_rexpr = function
  | RDeref e -> RDeref (simplify_rexpr e)
  | RAccess (e, n, l) -> RAccess (simplify_rexpr e, n, l)
  | RNotEmpty e -> RNotEmpty (simplify_rexpr e)
  | RHSlice { slice; rel_start; rel_end } ->
    let slice = simplify_rexpr slice in
    if (rel_start = 0 && rel_end = 0) then slice
    else RHSlice { slice; rel_start; rel_end }
  | RVSlice { slice; for_each; select } ->
    let slice = simplify_rexpr slice in
    if (for_each = select) then slice
    else RVSlice { slice; for_each; select }
  | RLShift (e, n) ->
    let e = simplify_rexpr e in
    if n = 0 then e else RLShift (e, n)
  | RRShift (e, n) ->
    let e = simplify_rexpr e in
    if n = 0 then e else RRShift (e, n)
  | RNot e ->
    (match simplify_rexpr e with
    | RWord z -> if Z.(z = zero) then RWord Z.one else RWord Z.zero
    | e -> RNot e)
  | RAnd (e, z) -> 
    if Z.(z = zero) then RWord Z.zero else RAnd (simplify_rexpr e, z)
  | RXor (e, z) ->
    let e = simplify_rexpr e in
    if Z.(z = zero) then e else RXor (e, z)
  | RLogAnd (e1, e2) ->
    let e1 = simplify_rexpr e1 and e2 = simplify_rexpr e2 in
    (match e1, e2 with
     | RWord z1, RWord z2 ->
       if Z.(z1 = zero) || Z.(z2 = zero) then RWord Z.zero
       else RWord Z.one
     | RWord z, e | e, RWord z ->
       if Z.(z = zero) then RWord Z.zero else e
     | _ -> RLogAnd (e1, e2))
  | RLogOr (e1, e2) ->
    let e1 = simplify_rexpr e1 and e2 = simplify_rexpr e2 in
    (match e1, e2 with
     | RWord z1, RWord z2 ->
       if Z.(z1 = one) || Z.(z2 = one) then RWord Z.one
       else RWord Z.zero
     | RWord z, e | e, RWord z ->
       if Z.(z = one) then RWord Z.one else e
     | _ -> RLogOr (e1, e2))
  | RArith (e, ar, z) ->
    let e = simplify_rexpr e in
    (match ar, Z.to_int z with
    | (Add | Sub), 0 -> e
    | Mul, 0 -> RWord Z.zero
    | (Mul | Div), 1 -> e
    | (Div | Mod), 0 -> failwith "wtfrudoin"
    | Mod, 1 -> RWord Z.zero
    | _ -> RArith (e, ar, z))
  | RCmp (e, cmp, z) ->
    if cmp = Geq && z = Z.zero then RWord Z.one
    else if cmp = Lt && z = Z.zero then RWord Z.zero
    else (match simplify_rexpr e with
    | RWord z' -> if cmp_to_z cmp z' z then RWord Z.one else RWord Z.zero
    | e -> RCmp (e, cmp, z))
  | e -> e

(** Decision trees *)

type lhs = LWord of word | LName of name
type dtree =
  | Fail
  | Success of (rexpr Name.Map.t * Syntax.expr)
  | Switch of {
      expr : rexpr ;
      cases : (lhs * dtree) list ;
      default : dtree option ;
    }

let switch ~default ~expr cases =
  let cases = List.filter (fun (_, tree) -> Some tree <> default) cases in
  let all_same_tree ~default l =
    let l = CCList.cons_maybe default @@ List.map snd cases in
    match l with
    | [] -> None
    | dt :: t ->
      if List.for_all (fun dt' -> dt = dt') t then
        Some dt
      else
        None
  in
  match cases, default with
  | [], None -> assert false
  | [_, x], None | [], Some x -> x
  | _ ->
    begin match all_same_tree ~default cases with
      | Some dt -> dt
      | None -> 
        Switch { default ; expr ; cases }
    end

module Dtree = struct

  module M = struct
    let lhs_equal lh1 lh2 = match lh1, lh2 with
      | LWord z1, LWord z2 -> Z.equal z1 z2
      | LName n1, LName n2 -> n1 = n2
      | _ -> assert false

    type t = dtree
    let rec equal x y = match x, y with
      | Fail, Fail -> true
      | Success (b1, e1), Success (b2, e2) ->
        Name.Map.equal (=) b1 b2 && e1 = e2
      | Switch s1, Switch s2 ->
        s1.expr = s2.expr &&
        CCEqual.option equal s1.default s2.default &&
        CCEqual.(list (pair lhs_equal equal)) s1.cases s2.cases
      | _, _ -> false
    let rec hash = function
      | Fail -> 0
      | Success (b, e) ->
        Hashtbl.hash (Name.Map.to_list b, e)
      | Switch s ->
        Hashtbl.hash
          (s.expr, List.map (CCPair.map_snd hash) s.cases,
           CCOption.map hash s.default)
  end
  module Tbl = Hashtbl.Make(M)
  
  module Hashcons = struct
    type 'a t = 'a Tbl.t
    let make () = Tbl.create 17
    let get = Tbl.find_opt
    let hashcons tbl x =
      match Tbl.find_opt tbl x with
      | Some x' -> x'
      | None -> Tbl.add tbl x x ; x
  end

  let share dt =
    let rec aux hcons dt =
      match Hashcons.get hcons dt with
      | Some dt -> dt
      | None ->
        let dt = match dt with
          | Fail | Success _ -> dt
          | Switch {expr; cases; default} ->
            let cases = List.map (fun (w, dt) -> (w, aux hcons dt)) cases in
            let default = CCOption.map (aux hcons) default in
            Switch {expr; cases; default}
        in
        Hashcons.hashcons hcons dt
    in
    aux (Hashcons.make ()) dt

  (* Sharing-aware fold on decision trees *)
  let fold ~f_switch ~f_success ~f_fail t =
    let tbl = Hashcons.make () in
    let rec aux dt =
      match Hashcons.get tbl dt with
      | Some x -> x
      | None ->
        let acc = match dt with
          | Switch {expr; cases; default} ->
            let cases = List.map (fun (w, dt) -> (w, aux dt)) cases in
            let default = CCOption.map aux default in
            f_switch expr cases default
          | Fail -> f_fail ()
          | Success suc -> f_success suc
        in
        Tbl.add tbl dt acc;
        acc
    in
    aux t
end

(** Printers *)

let rec pp_rexpr fmt re =
  let pp_hex fmt z = Fmt.pf fmt "%s" @@ Z.format "%x" z in
  match re with
  | RVar(x) -> Fmt.pf fmt "%s" x
  | RWord(z) -> Fmt.pf fmt "%a" pp_hex z
  | RDeref(e) -> Fmt.pf fmt "*%a" pp_rexpr_paren e
  (*| RAccess(e, l, 1) -> Fmt.pf fmt "%a[%d]" pp_rexpr_paren e l
  | RAccess(e, l, h) -> Fmt.pf fmt "%a[%d,%d]" pp_rexpr_paren e l h*)
  | RAccess(e, l, _) -> Fmt.pf fmt "%a.%d" pp_rexpr_paren e l
  | RNotEmpty e -> Fmt.pf fmt "%a ?" pp_rexpr_paren e
  | RHSlice { slice; rel_start; rel_end } ->
    Fmt.pf fmt "%a[[%d; %d]]" pp_rexpr_paren slice rel_start rel_end
  | RVSlice { slice; for_each; select } ->
    Fmt.pf fmt "%a[[%d / %d]]" pp_rexpr_paren slice for_each select
  | RLShift(e, n) -> Fmt.pf fmt "%a lsh %d" pp_rexpr_paren e n
  | RRShift(e, n) -> Fmt.pf fmt "%a rsh %d" pp_rexpr_paren e n
  | RAnd (e, w) -> Fmt.pf fmt "%a & %a" pp_rexpr_paren e pp_hex w
  | RNot e -> Fmt.pf fmt "! %a" pp_rexpr_paren e
  | RXor (e, w) -> Fmt.pf fmt "%a ^ %a" pp_rexpr_paren e pp_hex w
  | RLogAnd (e1, e2) -> Fmt.pf fmt "%a && %a" pp_rexpr_paren e1 pp_rexpr_paren e2
  | RLogOr (e1, e2) -> Fmt.pf fmt "%a || %a" pp_rexpr_paren e1 pp_rexpr_paren e2
  | RCmp (e, cmp, z) ->
    Fmt.pf fmt "%a %s %a" pp_rexpr_paren e (cmp_to_str cmp) pp_hex z
  | RArith (e, ar, z) ->
    Fmt.pf fmt "%a %s %a" pp_rexpr_paren e (arith_to_str ar) pp_hex z
and pp_rexpr_paren fmt re = match re with
  | RAccess _ | RHSlice _ | RVSlice _ | RNotEmpty _
  | RLShift _ | RRShift _ | RAnd _ | RXor _
  | RLogAnd _ | RLogOr _ | RArith _ -> Fmt.parens pp_rexpr fmt re
  | RVar _ | RWord _ | RDeref _ | RNot _ | RCmp _ -> pp_rexpr fmt re

let rec pp_dtree fmt dt = match dt with
  | Fail -> Fmt.pf fmt "fail"
  | Success(bindings, e) ->
    let pp_bindings fmt = 
      if Name.Map.is_empty bindings then ()
      else
        let f fmt (n,e) =
          Fmt.pf fmt "@[<2>let %a =@ %a@ in@]@ " Name.pp n pp_rexpr e
        in
        Fmt.pf fmt "%a"
          (Fmt.iter_bindings ~sep:Fmt.nop Name.Map.iter f)
          bindings
    in
    Fmt.pf fmt "@[<hv>%t@[success(%a)@]@]" pp_bindings Syntax.pp_expr e
  | Switch {expr; cases; default} ->
    let pp_lhs fmt = function
      | LWord z -> Fmt.pf fmt "%a" Z.pp_print z
      (*| LRange (z1,z2) -> Fmt.pf fmt "%a-%a" Z.pp_print z1 Z.pp_print z2*)
      | LName n -> Fmt.pf fmt "%s" n
    in
    let pp_case fmt (lhs, dt) =
      Fmt.pf fmt "@[<2>| %a ->@ %a@]" pp_lhs lhs pp_dtree dt
    in
    let pp_default fmt dt =
      Fmt.pf fmt "@[<2>| T ->@ %a@]" pp_dtree dt
    in 
    Fmt.pf fmt "@[<v 2>switch(%a) {@,%a%a@]@,}"
      pp_rexpr expr
      (Fmt.list ~sep:Fmt.cut pp_case) cases
      Fmt.(option (cut ++ pp_default)) default

(** Dot output *)
module Dot = struct
  type node =
    | Switch of rexpr * (int * lhs option) list
    | Success of (rexpr Name.Map.t * Syntax.expr)
    | Unreachable
  type edge = int * lhs option
    
  let as_dot_graph t =
    let r = ref 0 in
    let new_id () = incr r; !r in
    let vertices : (int * node) list ref = ref [] in
    let edges : (int * int, edge list) Hashtbl.t = Hashtbl.create 17 in
    let add_vertex x = vertices := x :: !vertices in
    let add_edge = CCHashtbl.add_list edges in
    let f_fail () =
      let id = new_id () in
      add_vertex (id, Unreachable);
      id
    and f_success b =
      let id = new_id () in
      add_vertex (id, Success b);
      id
    and f_switch r cases default =
      let id = new_id () in
      let all_cases =
        (match default with None -> [] | Some dti -> [dti, 0, None])
        @ List.mapi (fun i (w,dti) -> dti, (i+1), Some w) cases
      in
      add_vertex (id, Switch (r, List.map (fun (_,i,w) -> (i,w)) all_cases));
      List.iter (fun (dti, i, w) -> add_edge (id, dti) (i, w)) all_cases;
      id
    in
    let _ = Dtree.fold ~f_switch ~f_success ~f_fail t in
    (List.rev !vertices, edges)

  let pp_lhs fmt = function
    | LWord z -> Fmt.pf fmt "%a" Z.pp_print z
    (*| LRange (z1,z2) when Z.equal z1 z2 ->
      Fmt.pf fmt "%a" Z.pp_print z1
    | LRange (z1,z2) -> Fmt.pf fmt "%a-%a" Z.pp_print z1 Z.pp_print z2*)
    | LName n -> Fmt.pf fmt "%s" n

  let pp_case fmt = function
    | None -> Fmt.pf fmt "_"
    | Some tag -> Fmt.pf fmt "%a" pp_lhs tag

  let pp_bindings fmt bindings = 
    let f fmt (n,e) =
      Fmt.pf fmt {|%a = %a\l|} Name.pp n pp_rexpr e
    in
    Fmt.pf fmt "%a"
      (Fmt.iter_bindings ~sep:Fmt.nop Name.Map.iter f)
      bindings
          
  let pp_vertice ppf (id, node) =
    let attrs =
      match node with
      | Success (bindings, expr) ->
        let label =
          if Name.Map.is_empty bindings then
            Fmt.str "{%a}" Syntax.pp_expr expr
          else
            Fmt.str "{%a|%a}" pp_bindings bindings Syntax.pp_expr expr
        in
        [ ("label", label);
          ("style","filled");
          ("fillcolor","olivedrab2");
          ("shape", "Mrecord") ]
      | Unreachable -> 
        [ ("label", "Unreachable");
          ("style","filled");
          ("fillcolor","red");
          ("shape", "Mrecord") ]
      | Switch (r, cases) ->
        let f fmt (i,x) = Fmt.pf fmt "<%i> %a" i pp_case x in
        let label = Fmt.str "{ %a | { %a } }"
            pp_rexpr r Fmt.(list ~sep:(any " | ") f) cases
        in
        [ ("label", label); ("shape","record") ]
    in
    let pp_attr ppf (field, v) = Fmt.pf ppf {|%s="%s"|} field v in
    Fmt.pf ppf "%d [%a];" id (Fmt.list ~sep:(Fmt.any ", ") pp_attr) attrs

  let pp_edges ppf ((id, id'), tags) =
    let attrs = [] in
    let pp_attr ppf (field, v) = Fmt.pf ppf "%s=%S" field v in
    List.iter (fun (i, tag) -> 
        Fmt.pf ppf {|%d:%i:e -> %d:w [%a];|}
          id
          i
          id'
          (Fmt.list ~sep:(Fmt.any ", ") pp_attr)
          attrs
      ) tags

  let pp ppf t =
    let vertices, edges = as_dot_graph t in
    Fmt.pf ppf {|@[<v2>digraph G {@,concentrate=true;rankdir="LR";@,%a@,@,%a@,}@]|}
      (Fmt.list ~sep:Fmt.cut pp_vertice)
      vertices
      (Fmt.hashtbl ~sep:Fmt.cut pp_edges)
      edges
end

let pp_dot = Dot.pp
