type word_spec = { width: int; values: ValueSet.t }
type sum_spec = (Name.t * (Name.t list)) list

type 'a ty =
  | Type of 'a * Name.t list
  | Var of Name.t
  | Word of word_spec
  | Ref of Name.t
  | Product of Name.t list
  | Sum of sum_spec

type t = Name.t ty

type empty = | 
type monomorph = empty ty

type evaled_env = monomorph Name.Map.t

let rec unroll_vars (env : evaled_env) : monomorph -> monomorph = function
  | Var sym -> (match Name.Map.find_opt sym env with
    | Some ty -> unroll_vars env ty
    | None -> Var sym)
  | ty -> ty
let find_and_unroll sym (env : evaled_env) =
  unroll_vars env (Name.Map.find sym env)


type slot =
  | Lam of {
      parameters : Name.t list ;
      body : monomorph ;
      body_env : evaled_env ;
    }
  | TyVar of monomorph
type env = slot Name.Map.t

let address = Word { width = 64; values = ValueSet.non_zero 64 }

exception Unbound_type of Name.t

let rec cardinal (env : evaled_env) sym = (* TODO mostly wrong *)
  let addr_card = Z.(pow (one + one) 64) in
  match Name.Map.find sym env with
  | Var _ -> addr_card
  | Word { width; _ } -> Z.(pow (one + one) width)
  | Ref _ -> addr_card
  | Product l ->
    List.fold_left
      (fun c x -> Z.mul c @@ cardinal env x)
      Z.one l
  | Sum l ->
    List.fold_left
      (fun c (_, l) -> Z.add c @@
        List.fold_left (fun c x -> Z.mul c @@ cardinal env x) Z.one l)
      Z.zero l
  | Type _ -> .
  (* TODO: Refine this with proper treatment of recursive types *)

let min_width env t = Z.log2up @@ cardinal env t

(** Printers **)

let rec pp ppf fmt = function
  | Type (s, []) ->
    Fmt.pf fmt "%a" ppf s
  | Type (s, args) ->
    Fmt.pf fmt "%a(%a)" ppf s Fmt.(list ~sep:(any ",@ ") @@ Name.pp) args
  | Var x -> Name.pp fmt x
  | Word { width; values } ->
    Fmt.pf fmt "@[<2>bits<%i>%a@]" width (valueset width) values
  | Ref ty -> Fmt.pf fmt "&%s" ty
  | Product [] -> Fmt.pf fmt "()"
  | Product l -> 
    Fmt.pf fmt "<@[<hov>%a@]>"
      Fmt.(list ~sep:(any ";@ ") @@ Name.pp)
      l
  | Sum l ->
    let pp_constructor fmt = function
      | n, [] -> Name.pp fmt n
      | n, l -> Fmt.pf fmt "%a(%a)" Name.pp n (Fmt.list ~sep:Fmt.comma @@ Name.pp) l
    in
    Fmt.pf fmt "@[<hov>%a@]" Fmt.(list ~sep:(any "@ + ") pp_constructor) l

and valueset width fmt values =
  Fmt.pf fmt " (%a)" ValueSet.pp values

and pp_with_parens ppf fmt t =
  match t with
  | Word _ | Type (_, []) | Sum ([]|[_]) | Product ([]|[_]) | Var _ ->
    pp ppf fmt t
  | Ref _ | Sum _ | Product _ | Type (_, _) (*| Mu _*) ->
    Fmt.parens (pp ppf) fmt t

let pp_evaled fmt (x : monomorph) = pp Fmt.nop fmt x

let pp_evaled_env fmt env =
  Fmt.pf fmt "{ %a }@."
    Fmt.(list ~sep:(any " ;; @,") (pair ~sep:(any "->") Name.pp pp_evaled))
    (Name.Map.bindings env)

(** Error handling *)
let prepare_error = function
  | Unbound_type ty ->
    Some (Report.errorf
            "@[<v 2> Unbound type@;<1 2>%a@]"
            Name.pp ty)
  | _ -> None
let () = Report.register_report_of_exn prepare_error
